require 'fileutils'

module ServiceLayerRest

    #TODO: Generell Fehlverhalten wenn to oder from in tel/teldet null oder 1899... ist?
    #TODO: Fake für overlapping states [US-703]
    #TODO: Allgemeine Fake-Daten
  
    # MobileServiceController
    class MyPepController < ApplicationController
        include Base
        include Dll

        def get_detail_plan
            invalid_key = check_params([:fct_code, :from, :to], [:planarea_id, :employee_id])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?
        
            render get_detail_plan_local(
                params[:fct_code],
                params[:planarea_id],
                params[:employee_id],
                params[:from],
                params[:to]
            )
        end

        def get_detail_plan_local(fct_code, planarea_id, employee_id, from, to, for_timesheet = false)
            begin
                time_marker = start_timer
                response_string, string_pointer = ServiceLayerDll.get_detail_plan(
                    current_user.as_json,
                    fct_code,
                    planarea_id.to_i,
                    employee_id.to_i,
                    from,
                    to
                )
                stop_timer(time_marker, "Render calling DLL")
                ServiceLayerDll.free_string(string_pointer)

                time_marker = start_timer
                detail_plan, state = extract_data_and_state_from_response_string(response_string)
                stop_timer(time_marker, "Render making object from hash")

                time_marker = start_timer
                if state != Rack::Utils.status_code(:ok) || detail_plan.nil?
                    return parse_response_string_for_render(response_string)
                end
                
                [Constants::MIT_LIST_KEY, Constants::FERIENTAGE_KEY, Constants::PLANBEREICHE_KEY].each do |key| 
                    if detail_plan[key].nil? 
                        return plan_error(Constants::FROM_DETAIL_PLAN, Constants::INVALID_DATA)
                    end
                end
                
                mit_list = detail_plan[Constants::MIT_LIST_KEY]
                
                # Determine Planbereiche
                planbereiche = {}
                if detail_plan[Constants::PLANBEREICHE_KEY].present? && detail_plan[Constants::PLANBEREICHE_KEY][Constants::ITEMS_KEY].present?
                    detail_plan[Constants::PLANBEREICHE_KEY][Constants::ITEMS_KEY].compact.each { |planbereich| planbereiche[planbereich[Constants::ID_STATION_KEY]] = planbereich[Constants::PLANBEREICH_KEY] }
                end

                # Determine Mitarbeiterliste
                employeeArray = mit_list[Constants::ITEMS_KEY].compact
                if employeeArray.blank?
                    return plan_error(Constants::FROM_DETAIL_PLAN, Constants::EMPTY_EMPLOYEES)
                end

                # Determine Mitarbeiter and TKS-Array
                employee = nil
                tksArray = nil
                for employee in employeeArray
                    if employee[Constants::ID_SERVER_KEY] == employee_id.to_i
                        tksArray = employee[Constants::TKS_LIST_KEY]
                        break
                    end
                end

                # Check determined values from above
                if employee.nil? || tksArray.nil?
                    return plan_error(Constants::FROM_DETAIL_PLAN, Constants::EMPTY_DAILY_DATA)
                end
                tksItemArray = tksArray[Constants::ITEMS_KEY]
                if tksItemArray.nil?
                    return plan_error(Constants::FROM_DETAIL_PLAN, Constants::EMPTY_DAILY_DATA)
                end

                # Filter out null values
                tksItemArray = tksItemArray.compact
                
                smallest_change = nil
                dailyDataObjectArray = {}

                # Iterate through the TKS-Array
                for tksItem in tksItemArray
                    next if tksItem.nil? || tksItem[Constants::DATUM_KEY].nil?
                    begin
                        tksItemDate = tksItem[Constants::DATUM_KEY].to_time
                    rescue
                        next
                    end

                    # Look for the last change/creation of the plan
                    if tksItem[Constants::ERSTELLT_AM_KEY].present? && tksItem[Constants::ERSTELLT_AM_KEY] != Constants::DATE_ZERO
                        smallest_change = tksItem[Constants::ERSTELLT_AM_KEY].to_time if (smallest_change.nil?) || (smallest_change > tksItem[Constants::ERSTELLT_AM_KEY].to_time)
                    end
                    if tksItem[Constants::GEANDERT_AM_KEY].present? && tksItem[Constants::GEANDERT_AM_KEY] != Constants::DATE_ZERO
                        smallest_change = tksItem[Constants::GEANDERT_AM_KEY].to_time if (smallest_change.nil? || smallest_change > tksItem[Constants::GEANDERT_AM_KEY].to_time)
                    end

                    # Generate the daily data object and fill it with standard data
                    dailyDataObject = create_empty_daily_data_object tksItem
                    # Add the daily data object to the Array
                    dailyDataObjectArray[tksItemDate.strftime(Constants::DATE_FORMAT)] = dailyDataObject

                    planstatus = tksItem[Constants::PLANSTATUS_KEY]
                    states_overlapping_array = []

                    # Check the TEL-Array
                    if tksItem[Constants::TEL_LIST_SERVER_KEY].present? && tksItem[Constants::TEL_LIST_SERVER_KEY][Constants::ITEMS_KEY].present?
                        tel_items = tksItem[Constants::TEL_LIST_SERVER_KEY][Constants::ITEMS_KEY].compact
                        next if !tel_items.length

                        # Add tel_list for timesheet
                        if for_timesheet
                            dailyDataObject[Constants::TEL_LIST_CLIENT_KEY] = tel_items
                        end

                        dailyDataObject[Constants::FULL_DAY_STATE_KEY] = "" if check_for_data_in_tks_item(tksItem)
                        
                        # Fill the Left- & Right-Warnings of the daily data object
                        dailyDataObject[Constants::WARNING_LEFT_KEY] = []
                        dailyDataObject[Constants::WARNING_RIGHT_KEY] = []
                        # TODO: Remove Fake Data and add Live Data for Warnings [US-705]
                        add_warnings(tksItemDate, dailyDataObject)

                        planareas = []

                        # Iterate through the Tel-Array
                        for tel_item in tel_items
                            next if tel_item.nil? || tel_item[Constants::DATUM_KEY].nil? || tel_item[Constants::TAGESELEMENT_KATEGORIE_KEY].nil?

                            #TODO: Could also check anzahl property from the tel_item! (anzahl_TELDET)
                            next if tel_item[Constants::TEL_DET_LIST_KEY].nil? || tel_item[Constants::TEL_DET_LIST_KEY][Constants::ITEMS_KEY].nil?
                            teldet_items = tel_item[Constants::TEL_DET_LIST_KEY][Constants::ITEMS_KEY].compact
                            next if teldet_items.length == 0

                            # Check the type of the state
                            telItem = check_state_type(tel_item)

                            # Fill up the (unique) planareas for the daily data object
                            planareas.push(planbereiche[telItem[Constants::PLANAREA_ID_KEY]]) if planbereiche[telItem[Constants::PLANAREA_ID_KEY]].present? && !planareas.include?(planbereiche[telItem[Constants::PLANAREA_ID_KEY]])

                            # Sort TEL-Element by his category and push it into the right Array
                            if telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_SERVICE
                                dailyDataObject[Constants::SERVICES_KEY].push({Constants::TYPE_KEY => Constants::TYPE_SERVICE, Constants::ITEM_KEY => telItem})
                            elsif telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STANDBY || telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STANDBY_OPERATION
                                dailyDataObject[Constants::STANDBYS_KEY].push({Constants::TYPE_KEY => Constants::TYPE_STANDBY, Constants::ITEM_KEY => telItem})
                            elsif telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STATE
                                # Sort state types
                                if telItem[Constants::STATE_TYPE_KEY] == Constants::STATE_HOURLY
                                    dailyDataObject[Constants::STATES_KEY].push({Constants::TYPE_KEY => Constants::TYPE_STATE, Constants::ITEM_KEY => telItem})
                                elsif telItem[Constants::STATE_TYPE_KEY] == Constants::STATE_HOURLY_OVERLAPPING
                                    dailyDataObject[Constants::STATES_KEY].push({Constants::TYPE_KEY => Constants::TYPE_STATE, Constants::ITEM_KEY => telItem})
                                    states_overlapping_array.push(telItem)
                                elsif telItem[Constants::STATE_TYPE_KEY] == Constants::STATE_FULL && dailyDataObject[Constants::FULL_DAY_STATE_KEY].empty?
                                    dailyDataObject[Constants::FULL_DAY_STATE_KEY] = telItem[Constants::REFERENZ_TAGESDATEN_KEY] && telItem[Constants::REFERENZ_TAGESDATEN_KEY][Constants::BEZEICHNUNG_KEY] ? telItem[Constants::REFERENZ_TAGESDATEN_KEY][Constants::BEZEICHNUNG_KEY] : "N/A"
                                end
                            end

                            # Set From & To & Overnight Values
                            from = telItem[Constants::VON_KEY]
                            to = telItem[Constants::BIS_KEY]

                            # Check From & To
                            begin
                                dailyDataObject[Constants::DAY_FROM_KEY] = from.to_time if dailyDataObject[Constants::DAY_FROM_KEY] == "" && !from.blank?
                            rescue
                                # Silently ignore errors
                            end

                            begin
                                dailyDataObject[Constants::DAY_TO_KEY] = to.to_time unless to.blank?
                                dailyDataObject[Constants::OVERNIGHT_KEY] = true if !dailyDataObject[Constants::OVERNIGHT_KEY] && to.to_date == tksItemDate.next_day.beginning_of_day.to_date
                            rescue
                                # Silently ignore errors
                            end
                        end
                    end

                    # Add found planareas to the daily data object
                    dailyDataObject[Constants::PLANAREAS_KEY] = planareas
                    
                    # Add color infos to the daily data object
                    if tksItem[Constants::FARBINFO_ITEMLIST_KEY].present? && tksItem[Constants::FARBINFO_ITEMLIST_KEY][Constants::ITEMS_KEY].present?
                        dailyDataObject[Constants::COLOR_INFOS_KEY] = tksItem[Constants::FARBINFO_ITEMLIST_KEY][Constants::ITEMS_KEY].compact
                    end

                    # Add ferien to the daily data object
                    if detail_plan[Constants::FERIENTAGE_KEY].present? && detail_plan[Constants::FERIENTAGE_KEY][Constants::ITEMS_KEY].present?
                        ferien = detail_plan[Constants::FERIENTAGE_KEY][Constants::ITEMS_KEY].compact.find_all { |ferientag| ferientag[Constants::DATUM_KEY] == tksItem[Constants::DATUM_KEY] }
                        dailyDataObject[Constants::FERIEN_KEY] = ferien if ferien.present?
                    end

                    # Add plan state to the daily data object
                    dailyDataObject[Constants::PLANSTATUS_KEY] = planstatus
                end

                # Determine the current date of the given period from the params
                asked_for_date = nil
                begin
                    asked_for_date = params[:from].to_date + 7      # Need to add 7 days to get into the month
                rescue Exception => e
                    begin
                        asked_for_date = params[:to].to_date - 7    # Need to substract 7 days to get into the month
                    rescue Exception => e
                        # Silently ignore errors
                    end
                end

                # Determine the current header (& month confirmation object) for the given period
                if !for_timesheet
                    current_header, bHeaderBestaetigt, planNr = determine_header_from_employee(employee, asked_for_date)
                else
                    current_header = {}
                    bHeaderBestaetigt = -1
                    planNr = -1
                end

                informations = {}
                # We got the date (== month) we asked for and compare its month to the current month, to be sure to have the current header, 
                #    otherwise leave the informations empty
                if for_timesheet == false && asked_for_date != nil && asked_for_date.month == Time.now.month && current_header != {}
                    informations = determine_informations(employee, tksItemArray, asked_for_date, current_header, smallest_change)
                end

                data = {
                    Constants::DAILY_DATA_KEY => dailyDataObjectArray,
                    Constants::PLANBEREICHE_KEY => planbereiche,
                    Constants::INFORMATIONS_KEY => informations,
                    Constants::HEADER_KEY => current_header,
                        Constants::MONTH_CONFIRMATION_KEY => {
                        Constants::CONFIRMED_MONTH_KEY => bHeaderBestaetigt,
                    Constants::PLAN_NR_CLIENT_KEY => planNr
                    }
                }

                stop_timer(time_marker, "Render iterating & sorting")

                return create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, data)
            rescue Exception => e
                return parse_exception_for_render(e)
            end
        end

        def get_year_plan
            invalid_key = check_params([:fct_code, :from, :to], [:planarea_id, :employee_id])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?
            
            begin
                time_marker = start_timer
                employee_id = params[:employee_id]
                response_string, string_pointer = ServiceLayerDll.get_year_plan(
                    current_user.as_json,
                    params[:fct_code],
                    params[:planarea_id].to_i,
                    employee_id.to_i,
                    params[:from],
                    params[:to]
                )
                stop_timer(time_marker, "Render calling DLL")
                ServiceLayerDll.free_string(string_pointer)

                time_marker = start_timer
                year_plan, state = extract_data_and_state_from_response_string(response_string)
                stop_timer(time_marker, "Render making object from hash")

                time_marker = start_timer
                if state != Rack::Utils.status_code(:ok) || year_plan.nil?
                    render parse_response_string_for_render(response_string) 
                    return
                end

                [Constants::MIT_LIST_KEY, Constants::FERIENTAGE_KEY].each do |key| 
                    if year_plan[key].nil? 
                        render_plan_error(Constants::FROM_YEAR_PLAN, Constants::INVALID_DATA)
                        return
                    end
                end

                mit_list = year_plan[Constants::MIT_LIST_KEY]

                # Determine Mitarbeiterliste
                employeeArray = mit_list[Constants::ITEMS_KEY].compact
                if employeeArray.blank?
                    render_plan_error(Constants::FROM_YEAR_PLAN, Constants::EMPTY_EMPLOYEES)
                    return
                end

                # Determine Mitarbeiter and TKS-Array
                employee = nil
                tksArray = nil
                for employee in employeeArray
                    if employee[Constants::ID_SERVER_KEY] == employee_id.to_i
                        tksArray = employee[Constants::TKS_LIST_KEY]
                        break
                    end
                end
                
                # Check determined values from above
                if employee.nil? || tksArray.nil?
                    render_plan_error(Constants::FROM_YEAR_PLAN, Constants::EMPTY_DAILY_DATA)
                    return
                end
                tksItemArray = tksArray[Constants::ITEMS_KEY]
                if tksItemArray.nil?
                    render_plan_error(Constants::FROM_YEAR_PLAN, Constants::EMPTY_DAILY_DATA)
                    return
                end

                # Filter out null values
                tksItemArray = tksItemArray.compact
                
                dailyDataObjectArray = {}

                # Iterate through the TKS-Array
                for tksItem in tksItemArray
                    next if tksItem.nil? || tksItem[Constants::DATUM_KEY].nil?
                    begin
                        tksItemDate = tksItem[Constants::DATUM_KEY].to_time
                    rescue
                        next
                    end

                    # Generate the daily data object and fill it with standard data
                    dailyDataObject = create_empty_daily_data_object tksItem
                    # Add the daily data object to the Array
                    dailyDataObjectArray[tksItemDate.strftime(Constants::DATE_FORMAT)] = dailyDataObject

                    planstatus = tksItem[Constants::PLANSTATUS_KEY]
                    states_overlapping_array = []

                    # Check the TEL-Array
                    if tksItem[Constants::TEL_LIST_SERVER_KEY].present? && tksItem[Constants::TEL_LIST_SERVER_KEY][Constants::ITEMS_KEY].present?
                        tel_items = tksItem[Constants::TEL_LIST_SERVER_KEY][Constants::ITEMS_KEY].compact
                        next if !tel_items.length
                        
                        dailyDataObject[Constants::FULL_DAY_STATE_KEY] = "" if check_for_data_in_tks_item(tksItem, true)
                        
                        # Fill the Left- & Right-Warnings of the daily data object
                        dailyDataObject[Constants::WARNING_LEFT_KEY] = []
                        dailyDataObject[Constants::WARNING_RIGHT_KEY] = []
                        # TODO: Remove Fake Data and add Live Data for Warnings [US-705]
                        add_warnings(tksItemDate, dailyDataObject)

                        # Iterate through the Tel-Array
                        for tel_item in tel_items
                            next if tel_item.nil? || tel_item[Constants::DATUM_KEY].nil? || tel_item[Constants::TAGESELEMENT_KATEGORIE_KEY].nil?

                            next if tel_item[Constants::ANZAHL_TEL_DET_KEY] == 0

                            # Check the type of the state
                            telItem = check_state_type(tel_item)

                            # Sort TEL-Element by his category and push it into the right Array
                            if telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_SERVICE
                                dailyDataObject[Constants::SERVICES_KEY].push({Constants::TYPE_KEY => Constants::TYPE_SERVICE, Constants::ITEM_KEY => telItem})
                            elsif telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STANDBY || telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STANDBY_OPERATION
                                dailyDataObject[Constants::STANDBYS_KEY].push({Constants::TYPE_KEY => Constants::TYPE_STANDBY, Constants::ITEM_KEY => telItem})
                            elsif telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STATE
                                # Sort state types
                                if telItem[Constants::STATE_TYPE_KEY] == Constants::STATE_HOURLY
                                    dailyDataObject[Constants::STATES_KEY].push({Constants::TYPE_KEY => Constants::TYPE_STATE, Constants::ITEM_KEY => telItem})
                                elsif telItem[Constants::STATE_TYPE_KEY] == Constants::STATE_HOURLY_OVERLAPPING
                                    dailyDataObject[Constants::STATES_KEY].push({Constants::TYPE_KEY => Constants::TYPE_STATE, Constants::ITEM_KEY => telItem})
                                    states_overlapping_array.push(telItem)
                                elsif telItem[Constants::STATE_TYPE_KEY] == Constants::STATE_FULL && dailyDataObject[Constants::FULL_DAY_STATE_KEY].empty?
                                    dailyDataObject[Constants::FULL_DAY_STATE_KEY] = telItem[Constants::REFERENZ_TAGESDATEN_KEY] && telItem[Constants::REFERENZ_TAGESDATEN_KEY][Constants::BEZEICHNUNG_KEY] ? telItem[Constants::REFERENZ_TAGESDATEN_KEY][Constants::BEZEICHNUNG_KEY] : "N/A"
                                end
                            end

                            # Set From & To & Overnight Values
                            to = telItem[Constants::BIS_KEY]
                            begin
                                dailyDataObject[Constants::OVERNIGHT_KEY] = true if !dailyDataObject[Constants::OVERNIGHT_KEY] && to.to_date == tksItemDate.next_day.beginning_of_day.to_date
                            rescue
                                # Silently ignore errors
                            end
                        end
                    end

                    # Add color infos to the daily data object
                    if tksItem[Constants::FARBINFO_ITEMLIST_KEY].present? && tksItem[Constants::FARBINFO_ITEMLIST_KEY][Constants::ITEMS_KEY].present?
                        dailyDataObject[Constants::COLOR_INFOS_KEY] = tksItem[Constants::FARBINFO_ITEMLIST_KEY][Constants::ITEMS_KEY].compact
                    end

                    # Add ferien to the daily data object
                    if year_plan[Constants::FERIENTAGE_KEY].present? && year_plan[Constants::FERIENTAGE_KEY][Constants::ITEMS_KEY].present?
                        ferien = year_plan[Constants::FERIENTAGE_KEY][Constants::ITEMS_KEY].compact.find_all { |ferientag| ferientag[Constants::DATUM_KEY] == tksItem[Constants::DATUM_KEY] }
                        dailyDataObject[Constants::FERIEN_KEY] = ferien if ferien.present?
                    end

                    # Add plan state to the daily data object
                    dailyDataObject[Constants::PLANSTATUS_KEY] = planstatus
                end

                data = {
                    Constants::DAILY_DATA_KEY => dailyDataObjectArray,
                }

                stop_timer(time_marker, "Render iterating & sorting")

                render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, data)
            rescue Exception => e
                render parse_exception_for_render(e)
            end
        end

        # From & To are no parameters, cause informations are only for the current month
        def get_informations
            invalid_key = check_params([:fct_code], [:planarea_id, :employee_id])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?

            begin
                time_marker = start_timer
                now = Time.now
                employee_id = params[:employee_id]
                response_string, string_pointer = ServiceLayerDll.get_staff_scheduling_daily_data_all(
                    current_user.as_json,
                    params[:fct_code],
                    params[:planarea_id].to_i,
                    employee_id.to_i,
                    now.at_beginning_of_month.strftime('%d.%m.%Y'),
                    now.end_of_month.strftime('%d.%m.%Y')
                )
                stop_timer(time_marker, "Render calling DLL")
                ServiceLayerDll.free_string(string_pointer)

                time_marker = start_timer
                mit_list, state = extract_data_and_state_from_response_string(response_string)
                stop_timer(time_marker, "Render making object from hash")

                time_marker = start_timer
                if state != Rack::Utils.status_code(:ok) || mit_list.nil?
                    render parse_response_string_for_render(response_string)
                    return
                end
                if mit_list.nil? || mit_list.empty? 
                    render_plan_error(Constants::FROM_INFORMATIONS, Constants::INVALID_DATA)
                    return
                end
                
                # Determine Mitarbeiterliste
                employeeArray = mit_list[Constants::ITEMS_KEY].compact
                if employeeArray.blank?
                    render_plan_error(Constants::FROM_INFORMATIONS, Constants::EMPTY_EMPLOYEES)
                    return
                end

                # Determine Mitarbeiter and TKS-Array
                employee = nil
                tksArray = nil
                for employee in employeeArray
                    if employee[Constants::ID_SERVER_KEY] == employee_id.to_i
                        tksArray = employee[Constants::TKS_LIST_KEY]
                        break
                    end
                end

                # Check determined values from above
                if employee.nil? || tksArray.nil?
                    render_plan_error(Constants::FROM_INFORMATIONS, Constants::EMPTY_DAILY_DATA)
                    return
                end
                tksItemArray = tksArray[Constants::ITEMS_KEY]
                if tksItemArray.nil?
                    render_plan_error(Constants::FROM_INFORMATIONS, Constants::EMPTY_DAILY_DATA)
                    return
                end

                # Filter out null values
                tksItemArray = tksItemArray.compact
                
                smallest_change = nil
                
                # Iterate through the TKS-Array
                for tksItem in tksItemArray
                    next if tksItem.nil? || tksItem[Constants::DATUM_KEY].nil?
                    begin
                        tksItemDate = tksItem[Constants::DATUM_KEY].to_time
                    rescue
                        next
                    end

                    # Look for the last change/creation of the plan
                    if tksItem[Constants::ERSTELLT_AM_KEY].present? && tksItem[Constants::ERSTELLT_AM_KEY] != Constants::DATE_ZERO
                        smallest_change = tksItem[Constants::ERSTELLT_AM_KEY].to_time if (smallest_change.nil?) || (smallest_change > tksItem[Constants::ERSTELLT_AM_KEY].to_time)
                    end
                    if tksItem[Constants::GEANDERT_AM_KEY].present? && tksItem[Constants::GEANDERT_AM_KEY] != Constants::DATE_ZERO
                        smallest_change = tksItem[Constants::GEANDERT_AM_KEY].to_time if (smallest_change.nil? || smallest_change > tksItem[Constants::GEANDERT_AM_KEY].to_time)
                    end
                end

                # Determine the current header for the given period
                current_header, bHeaderBestaetigt, planNr = determine_header_from_employee(employee, now)
                
                informations = current_header != {} ? determine_informations(employee, tksItemArray, now, current_header, smallest_change) : {}

                data = {
                    Constants::INFORMATIONS_KEY => informations
                }

                stop_timer(time_marker, "Render iterating & sorting")

                render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, data)
            rescue Exception => e
                render parse_exception_for_render(e)
            end
        end

        def get_group_plan
            invalid_key = check_params([:fct_code, :from, :to], [:planarea_id, :employee_id])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?

            begin
                time_marker = start_timer
                planarea_id = params[:planarea_id].to_i
                response_string, string_pointer = ServiceLayerDll.get_group_plan(
                    current_user.as_json,
                    params[:fct_code],
                    planarea_id,
                    params[:employee_id].to_i,
                    params[:from],
                    params[:to]
                )
                stop_timer(time_marker, "Render calling DLL")
                ServiceLayerDll.free_string(string_pointer)

                time_marker = start_timer
                group_plan, state = extract_data_and_state_from_response_string(response_string)
                stop_timer(time_marker, "Render making object from hash")

                time_marker = start_timer
                if state != Rack::Utils.status_code(:ok) || group_plan.nil?
                    render parse_response_string_for_render(response_string)
                    return
                end
                
                [Constants::MIT_LIST_KEY, Constants::FERIENTAGE_KEY].each do |key| 
                    if group_plan[key].nil? 
                        render_plan_error(Constants::FROM_GROUP_PLAN, Constants::INVALID_DATA)
                        return
                    end
                end

                mit_list = group_plan[Constants::MIT_LIST_KEY]

                # Determine Mitarbeiterliste
                employeeArray = mit_list[Constants::ITEMS_KEY].compact
                if employeeArray.blank?
                    render_plan_error(Constants::FROM_GROUP_PLAN, Constants::EMPTY_EMPLOYEES)
                    return
                end
                
                employee_array = []

                # Determine Mitarbeiter and TKS-Array
                for employee in employeeArray
                    tksArray = employee[Constants::TKS_LIST_KEY]
                
                    # Check determined values from above
                    if employee.nil? || tksArray.nil?
                        render_plan_error(Constants::FROM_GROUP_PLAN, Constants::EMPTY_DAILY_DATA)
                        return
                    end
                    tksItemArray = tksArray[Constants::ITEMS_KEY]
                    if tksItemArray.nil?
                        render_plan_error(Constants::FROM_GROUP_PLAN, Constants::EMPTY_DAILY_DATA)
                        return
                    end

                    # Filter out null values
                    tksItemArray = tksItemArray.compact
                    
                    dailyDataArray = []

                    # Has stuff on different planareas
                    meat_all = false

                    # Iterate through the TKS-Array
                    for tksItem in tksItemArray
                        next if tksItem.nil? || tksItem[Constants::DATUM_KEY].nil?
                        begin
                            tksItemDate = tksItem[Constants::DATUM_KEY].to_time
                        rescue
                            next
                        end

                        # Generate the daily data object and fill it with standard data
                        dailyDataObject = create_empty_daily_data_object tksItem

                        planstatus = tksItem[Constants::PLANSTATUS_KEY]
                        states_overlapping_array = []

                        # Check the TEL-Array
                        if tksItem[Constants::TEL_LIST_SERVER_KEY].present? && tksItem[Constants::TEL_LIST_SERVER_KEY][Constants::ITEMS_KEY].present?
                            tel_items = tksItem[Constants::TEL_LIST_SERVER_KEY][Constants::ITEMS_KEY].compact
                            next if !tel_items.length

                            dailyDataObject[Constants::FULL_DAY_STATE_KEY] = "" if check_for_data_in_tks_item(tksItem, true)
                        
                            meat = false
                            matching = false

                            # Iterate through the Tel-Array
                            for tel_item in tel_items
                                # Filter out invalid planareas (cause backend gives me all)
                                next if tel_item.nil? || tel_item[Constants::DATUM_KEY].nil? || 
                                        tel_item[Constants::TAGESELEMENT_KATEGORIE_KEY].nil?                                        

                                next if tel_item[Constants::ANZAHL_TEL_DET_KEY] == 0

                                if tel_item[Constants::PLANAREA_ID_KEY] != planarea_id
                                    meat = true
                                    meat_all = true

                                    next
                                else
                                    matching = true
                                end

                                # Check the type of the state
                                telItem = check_state_type(tel_item)

                                # Sort TEL-Element by his category and push it into the right Array
                                if telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_SERVICE
                                    dailyDataObject[Constants::SERVICES_KEY].push({Constants::TYPE_KEY => Constants::TYPE_SERVICE, Constants::ITEM_KEY => telItem})
                                elsif telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STANDBY || telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STANDBY_OPERATION
                                    dailyDataObject[Constants::STANDBYS_KEY].push({Constants::TYPE_KEY => Constants::TYPE_STANDBY, Constants::ITEM_KEY => telItem})
                                elsif telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STATE
                                    # Sort state types
                                    print_kuerzel = check_print_kuerzel(telItem, employee[Constants::ID_SERVER_KEY])
                                    tel_item[Constants::KUERZEL_IST_KEY] = print_kuerzel if print_kuerzel != ""
                                    if telItem[Constants::STATE_TYPE_KEY] == Constants::STATE_HOURLY
                                        dailyDataObject[Constants::STATES_KEY].push({Constants::TYPE_KEY => Constants::TYPE_STATE, Constants::ITEM_KEY => telItem})
                                    elsif telItem[Constants::STATE_TYPE_KEY] == Constants::STATE_HOURLY_OVERLAPPING
                                        dailyDataObject[Constants::STATES_KEY].push({Constants::TYPE_KEY => Constants::TYPE_STATE, Constants::ITEM_KEY => telItem})
                                        states_overlapping_array.push(telItem)
                                    elsif telItem[Constants::STATE_TYPE_KEY] == Constants::STATE_FULL && dailyDataObject[Constants::FULL_DAY_STATE_KEY].empty?
                                        if print_kuerzel != ""
                                            dailyDataObject[Constants::FULL_DAY_STATE_KEY] = print_kuerzel
                                        else
                                            dailyDataObject[Constants::FULL_DAY_STATE_KEY] = telItem[Constants::REFERENZ_TAGESDATEN_KEY] && telItem[Constants::REFERENZ_TAGESDATEN_KEY][Constants::BEZEICHNUNG_KEY] ? telItem[Constants::REFERENZ_TAGESDATEN_KEY][Constants::BEZEICHNUNG_KEY] : "N/A"
                                        end
                                    end
                                end

                                # Set From & To & Overnight Values
                                to = telItem[Constants::BIS_KEY]
                                begin
                                    dailyDataObject[Constants::OVERNIGHT_KEY] = true if !dailyDataObject[Constants::OVERNIGHT_KEY] && to.to_date == tksItemDate.next_day.beginning_of_day.to_date
                                rescue
                                    # Silently ignore errors
                                end
                            end

                            dailyDataObject[Constants::FULL_DAY_STATE_KEY] = '-' if meat && !matching
                        end

                        # Add color infos to the daily data object
                        if tksItem[Constants::FARBINFO_ITEMLIST_KEY].present? && tksItem[Constants::FARBINFO_ITEMLIST_KEY][Constants::ITEMS_KEY].present?
                            dailyDataObject[Constants::COLOR_INFOS_KEY] = tksItem[Constants::FARBINFO_ITEMLIST_KEY][Constants::ITEMS_KEY].compact
                        end

                        # Add plan state to the daily data object
                        dailyDataObject[Constants::PLANSTATUS_KEY] = planstatus
                
                        # Add the daily data object to the Array
                        dailyDataObject[Constants::DATE_KEY] = tksItemDate.strftime(Constants::DATE_REVERSED_FORMAT)
                        dailyDataArray.push(dailyDataObject)
                    end

                    employee_array.push({
                        Constants::ID_CLIENT_KEY => employee[Constants::ID_SERVER_KEY],
                        Constants::FIRSTNAME_KEY => employee[Constants::VORNAME_KEY],
                        Constants::LASTNAME_KEY => employee[Constants::NACHNAME_KEY],
                        Constants::DAILY_DATA_ALT_KEY => dailyDataArray,
                        Constants::MEAT_KEY => meat_all
                    })
                end

                # Add ferien to the daily data object
                ferien = []
                if group_plan[Constants::FERIENTAGE_KEY].present? && group_plan[Constants::FERIENTAGE_KEY][Constants::ITEMS_KEY].present?
                    ferien = group_plan[Constants::FERIENTAGE_KEY][Constants::ITEMS_KEY].compact
                end

                data = {
                    Constants::MIT_LIST_KEY => employee_array,
                    Constants::FERIEN_KEY => ferien
                }

                stop_timer(time_marker, "Render iterating & sorting")

                render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, data)
            rescue Exception => e
                render parse_exception_for_render(e)
            end
        end

        def get_daily_data_details
            invalid_key = check_params([:fct_code, :date], [:planarea_id, :employee_id])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?

            begin
                time_marker = start_timer
                response_string, string_pointer = ServiceLayerDll.get_daily_data_details(
                    current_user.as_json,
                    params[:fct_code],
                    params[:planarea_id].to_i,
                    params[:employee_id].to_i,
                    params[:date]
                )
                stop_timer(time_marker, "Render calling DLL")
                ServiceLayerDll.free_string(string_pointer)

                time_marker = start_timer
                daily_data_details, state = extract_data_and_state_from_response_string(response_string)
                stop_timer(time_marker, "Render making object from hash")

                time_marker = start_timer
                if state != Rack::Utils.status_code(:ok) || daily_data_details.nil?
                    render parse_response_string_for_render(response_string)
                    return
                end

                [Constants::MIT_LIST_KEY, Constants::FERIENTAGE_KEY, Constants::PLANBEREICHE_KEY].each do |key| 
                    if daily_data_details[key].nil? 
                        render_plan_error(Constants::FROM_DETAILS, Constants::INVALID_DATA)
                        return
                    end
                end

                mit_list = daily_data_details[Constants::MIT_LIST_KEY]

                # Determine Planbereiche
                planbereiche = {}
                if daily_data_details[Constants::PLANBEREICHE_KEY].present? && daily_data_details[Constants::PLANBEREICHE_KEY][Constants::ITEMS_KEY].present?
                    daily_data_details[Constants::PLANBEREICHE_KEY][Constants::ITEMS_KEY].compact.each { |planbereich| planbereiche[planbereich[Constants::ID_STATION_KEY]] = planbereich[Constants::PLANBEREICH_KEY] }
                end

                # Determine Mitarbeiterliste
                employeeArray = mit_list[Constants::ITEMS_KEY].compact
                if employeeArray.blank?
                    render_plan_error(Constants::FROM_DETAILS, Constants::EMPTY_EMPLOYEES)
                    return
                end

                # Determine Mitarbeiter and TKS-Array
                employee = nil
                tksArray = nil
                if employeeArray.length == 1
                    employee = employeeArray[0]
                    tksArray = employee[Constants::TKS_LIST_KEY] if employee.present?
                end
                
                # Check determined values from above
                if employee.nil? || tksArray.nil?
                    render_plan_error(Constants::FROM_DETAILS, Constants::EMPTY_DAILY_DATA)
                    return
                end
                tksItemArray = tksArray[Constants::ITEMS_KEY]
                if tksItemArray.nil?
                    render_plan_error(Constants::FROM_DETAILS, Constants::EMPTY_DAILY_DATA)
                    return
                end

                # Filter out null values
                tksItemArray = tksItemArray.compact
                
                dailyDataObjectArray = {}

                # Iterate through the TKS-Array
                for tksItem in tksItemArray
                    next if tksItem.nil? || tksItem[Constants::DATUM_KEY].nil?
                    begin
                        tksItemDate = tksItem[Constants::DATUM_KEY].to_time
                    rescue
                        next
                    end

                    # Generate the daily data object and fill it with standard data
                    dailyDataObject = create_empty_daily_data_object tksItem
                    # Add the daily data object to the Array
                    dailyDataObjectArray[tksItemDate.strftime(Constants::DATE_FORMAT)] = dailyDataObject

                    planstatus = tksItem[Constants::PLANSTATUS_KEY]
                    states_overlapping_array = []

                    filtered_tel_list = []

                    # Check the TEL-Array
                    if tksItem[Constants::TEL_LIST_SERVER_KEY].present? && tksItem[Constants::TEL_LIST_SERVER_KEY][Constants::ITEMS_KEY].present?
                        tel_items = tksItem[Constants::TEL_LIST_SERVER_KEY][Constants::ITEMS_KEY].compact
                        next if !tel_items.length

                        dailyDataObject[Constants::FULL_DAY_STATE_KEY] = "" if check_for_data_in_tks_item(tksItem)

                        # Fill the Left- & Right-Warnings of the daily data object
                        dailyDataObject[Constants::WARNING_LEFT_KEY] = []
                        dailyDataObject[Constants::WARNING_RIGHT_KEY] = []
                        # TODO: Remove Fake Data and add Live Data for Warnings [US-705]
                        add_warnings(tksItemDate, dailyDataObject)

                        # Iterate through the Tel-Array
                        for tel_item in tel_items
                            next if tel_item.nil? || tel_item[Constants::DATUM_KEY].nil? || tel_item[Constants::TAGESELEMENT_KATEGORIE_KEY].nil?

                            next if tel_item[Constants::TEL_DET_LIST_KEY].nil? || tel_item[Constants::TEL_DET_LIST_KEY][Constants::ITEMS_KEY].nil?

                            teldet_items = tel_item[Constants::TEL_DET_LIST_KEY][Constants::ITEMS_KEY].compact
                            next if teldet_items.length == 0

                            # Check the type of the state
                            telItem = check_state_type(tel_item)

                            # Sort TEL-Element by his category and push it into the right Array
                            if telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STATE
                                # Sort state types
                                print_kuerzel = check_print_kuerzel(telItem, employee[Constants::ID_SERVER_KEY])
                                tel_item[Constants::KUERZEL_IST_KEY] = print_kuerzel if print_kuerzel != ""
                                
                                if telItem[Constants::STATE_TYPE_KEY] == Constants::STATE_HOURLY_OVERLAPPING
                                    states_overlapping_array.push(telItem)
                                elsif telItem[Constants::STATE_TYPE_KEY] == Constants::STATE_FULL && dailyDataObject[Constants::FULL_DAY_STATE_KEY].empty?
                                    if print_kuerzel != ""
                                        dailyDataObject[Constants::FULL_DAY_STATE_KEY] = print_kuerzel
                                    else
                                        dailyDataObject[Constants::FULL_DAY_STATE_KEY] = telItem[Constants::REFERENZ_TAGESDATEN_KEY] && telItem[Constants::REFERENZ_TAGESDATEN_KEY][Constants::BEZEICHNUNG_KEY] ? telItem[Constants::REFERENZ_TAGESDATEN_KEY][Constants::BEZEICHNUNG_KEY] : "N/A"
                                    end
                                end
                            end

                            # Sort type of TEL-Element
                            if telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_SERVICE
                                telItem[Constants::TYPE_KEY] = Constants::TYPE_SERVICE;
                                dailyDataObject[Constants::SERVICES_KEY].push(telItem)
                            elsif telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STANDBY || tksItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STANDBY_OPERATION
                                telItem[Constants::TYPE_KEY] = Constants::TYPE_STANDBY;
                                dailyDataObject[Constants::STANDBYS_KEY].push(telItem)
                            elsif telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STATE
                                telItem[Constants::TYPE_KEY] = Constants::TYPE_STATE;
                                dailyDataObject[Constants::STATES_KEY].push(telItem)
                            end    

                            # Add element to filtered_tel_list
                            filtered_tel_list.push(telItem) if telItem[Constants::STATE_TYPE_KEY] == Constants::STATE_HOURLY || telItem[Constants::STATE_TYPE_KEY] == "-"

                            # Set From & To & Overnight Values
                            for teldetItem in teldet_items
                                to = teldetItem[Constants::BIS_KEY]

                                # Better check teldetItem, what about breaks?
                                print_kuerzel = (telItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STATE) ? check_print_kuerzel(teldetItem, employee[Constants::ID_SERVER_KEY]) : ""
                                if print_kuerzel != ""
                                    teldetItem[Constants::KUERZEL_IST_KEY] = print_kuerzel 
                                    if teldetItem[Constants::REFERENZ_TAGESDATEN_KEY]
                                        teldetItem[Constants::REFERENZ_TAGESDATEN_KEY][Constants::BEZEICHNUNG_KEY] = nil
                                    end
                                end

                                begin
                                    dailyDataObject[Constants::OVERNIGHT_KEY] = true if !dailyDataObject[Constants::OVERNIGHT_KEY] && to.to_date == tksItemDate.next_day.beginning_of_day.to_date
                                rescue
                                    # Silently ignore errors
                                end
                            end
                        end

                        # Iterate through the Tel-Array
                        index_to_add = rand(0..filtered_tel_list.length - 1) # Fake
                        filtered_tel_list.each_with_index do |filtered_tel_item, index|
                            for state_overlapping_item in states_overlapping_array
                                # TODO: Adjust fake matching for hourly overlapping states to the tel_elements [US-703] ? Really? Bleibt das nciht so hier?
                                if (index == index_to_add)
                                    status_start = nil
                                    status_end = nil
                                    filtered_teldet_list = filtered_tel_list[index][Constants::TEL_DET_LIST_KEY][Constants::ITEMS_KEY].compact
                                    for filtered_tel_det_item in filtered_teldet_list
                                        from = filtered_tel_det_item[Constants::VON_KEY]
                                        to = filtered_tel_det_item[Constants::BIS_KEY]

                                        # Check From & To
                                        begin
                                            status_start = from.to_time if status_start.nil? && !from.blank?
                                        rescue
                                            # Silently ignore errors
                                        end

                                        begin
                                            status_end = to.to_time unless to.blank?
                                        rescue
                                            # Silently ignore errors
                                        end
                                    end
                                    if status_start.present? && status_end.present?
                                        print_kuerzel = check_print_kuerzel(filtered_tel_det_item, employee[Constants::ID_SERVER_KEY])
                                        if print_kuerzel != ""
                                            bezeichnung = print_kuerzel
                                        else
                                            bezeichnung = state_overlapping_item[Constants::REFERENZ_TAGESDATEN_KEY] && state_overlapping_item[Constants::REFERENZ_TAGESDATEN_KEY][Constants::BEZEICHNUNG_KEY] ? state_overlapping_item[Constants::REFERENZ_TAGESDATEN_KEY][Constants::BEZEICHNUNG_KEY] : "N/A"
                                        end
                                        if $fake_data
                                            filtered_tel_list[index][Constants::STATUS_HOUR_KEY].push({
                                                Constants::VON_KEY => status_start + 60 * 15,
                                                Constants::BIS_KEY => status_start + 60 * 30,
                                                Constants::NAME_KEY => bezeichnung,
                                            })
                                            filtered_tel_list[index][Constants::STATUS_HOUR_KEY].push({
                                                Constants::VON_KEY => status_start + 60 * 30,
                                                Constants::BIS_KEY => status_end - 60 * 25,
                                                Constants::NAME_KEY => bezeichnung,
                                            })
                                        else
                                            filtered_tel_list[index][Constants::STATUS_HOUR_KEY].push({
                                                Constants::VON_KEY => status_start + 60 * 5,
                                                Constants::BIS_KEY => status_end - 60 * 15,
                                                Constants::NAME_KEY => bezeichnung,
                                            })
                                        end
                                    end
                                end
                            end
                        end
                    end

                    # Add plan state to the daily data object
                    dailyDataObject[Constants::PLANSTATUS_KEY] = planstatus 

                    # Add filtered tel_list to the daily data object
                    dailyDataObject[Constants::TEL_LIST_CLIENT_KEY] = filtered_tel_list

                    # Add color infos to the daily data object
                    if tksItem[Constants::FARBINFO_ITEMLIST_KEY].present? && tksItem[Constants::FARBINFO_ITEMLIST_KEY][Constants::ITEMS_KEY].present?
                        dailyDataObject[Constants::COLOR_INFOS_KEY] = tksItem[Constants::FARBINFO_ITEMLIST_KEY][Constants::ITEMS_KEY].compact
                    end

                    # Add ferien to the daily data object
                    if daily_data_details[Constants::FERIENTAGE_KEY].present? && daily_data_details[Constants::FERIENTAGE_KEY][Constants::ITEMS_KEY].present?
                        ferien = daily_data_details[Constants::FERIENTAGE_KEY][Constants::ITEMS_KEY].compact.find_all { |ferientag| ferientag[Constants::DATUM_KEY] == tksItem[Constants::DATUM_KEY] }
                        dailyDataObject[Constants::FERIEN_KEY] = ferien if ferien.present?
                    end
                end

                data = {
                    Constants::DETAILS_KEY => dailyDataObjectArray,
                    Constants::PLANBEREICHE_KEY => planbereiche
                }

                stop_timer(time_marker, "Render iterating & sorting")

                render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, data)
            rescue Exception => e
                render parse_exception_for_render(e)
            end
        end

        def delete_daily_data_duty_requests
            invalid_key = check_params([:fct_code], [])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?

            checked_params = check_params_for_tlist(params, :add_tel_list_service, Constants::DELETE_SERVICES, true)
            return render checked_params unless checked_params == true
            checked_params = check_params_for_tlist(params, :add_tel_list_standby, Constants::DELETE_STANDBYS, true)
            return render checked_params unless checked_params == true
            checked_params = check_params_for_tlist(params, :add_tel_list_state, Constants::DELETE_STATES, true)
            return render checked_params unless checked_params == true

            begin
                response_string, string_pointer = ServiceLayerDll.delete_daily_data_duty_requests(
                    current_user.as_json,
                    params[:fct_code],
                    params[:add_tel_list_service].to_json,
                    params[:add_tel_list_standby].to_json,
                    params[:add_tel_list_state].to_json
                )
                ServiceLayerDll.free_string(string_pointer)
            
                render parse_response_string_for_render(response_string.force_encoding('utf-8'))
            rescue Exception => e
                render parse_exception_for_render(e)
            end
        end
        
        def get_definitions
            invalid_key = check_params([:fct_code, :date], [:planarea_id, :employee_id])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?

            begin
                time_marker = start_timer
                response_string, string_pointer = ServiceLayerDll.get_definitions(
                    current_user.as_json,
                    params[:fct_code],
                    params[:planarea_id].to_i,
                    params[:employee_id].to_i,
                    params[:date]
                )
                stop_timer(time_marker, "Render calling DLL")
                ServiceLayerDll.free_string(string_pointer)

                handle_get_definitions_response_string(response_string)
            rescue Exception => e
                render parse_exception_for_render(e)
            end
        end

        def get_definitions_for_datearray
            invalid_key = check_params([:fct_code, :datearray], [:planarea_id, :employee_id])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?

            begin
                time_marker = start_timer
                response_string, string_pointer = ServiceLayerDll.get_definitions_for_datearray(
                    current_user.as_json,
                    params[:fct_code],
                    params[:planarea_id].to_i,
                    params[:employee_id].to_i,
                    params[:datearray].to_s
                )
                stop_timer(time_marker, "Render calling DLL")
                ServiceLayerDll.free_string(string_pointer)

                handle_get_definitions_response_string(response_string)
            rescue Exception => e
                render parse_exception_for_render(e)
            end
        end

        def load_timesheet
            invalid_key = check_params([:fct_code, :from, :to], [:employee_id])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?

            return render bad_params_for_render(:withPlanData) unless params.key?(:withPlanData)
            withPlanData = params[:withPlanData] == "true" ? true : false

            employee_id = params[:employee_id]

            # TODO: US-715 - Fake data loading. Replace with loading from the database when ready. ALso adjust merging of data and use Constants!
            fake_timesheet = true
            if fake_timesheet
                fake_file_path = File.dirname(__FILE__)+"/my_pep_fake_timesheet.json"
                employee_file_path = File.dirname(__FILE__)+"/my_pep_timesheet_#{employee_id.to_i}.json"
                if !File.file?(employee_file_path)
                    FileUtils.cp(fake_file_path, employee_file_path)
                end
                employee_data = JSON.parse(File.read(employee_file_path))

                # Fill up the empty data
                if withPlanData
                    detail_plan = get_detail_plan_local(
                        params[:fct_code],
                        0,
                        employee_id.to_i,
                        params[:from],
                        params[:to],
                        true
                    )
                    
                    daily_data = nil
                    planareas = nil
                    begin
                        daily_data = detail_plan[:json][:data][Constants::DAILY_DATA_KEY]
                        planareas = detail_plan[:json][:data][Constants::PLANBEREICHE_KEY]
                    rescue Exception => e
                        render parse_exception_for_render(e)
                        return
                    end

                    if daily_data.nil? || planareas.nil? || daily_data.empty? || planareas.empty?
                        render_plan_error(Constants::FROM_TIMESHEET, Constants::INVALID_DATA)
                        return
                    end

                    # Fill up empty data with keys
                    (Date.parse(params[:from])..Date.parse(params[:to])).map do |date|
                        key = date.strftime(Constants::DATE_FORMAT)
                        employee_data[key] = [] if employee_data[key].nil?
                    end

                    ignore_days = []
                    not_in_time_days = []
                    # Filter out days, which could be ignored without going into the deep
                    employee_data.each do |key, value| 
                        next if value.nil? || value.empty?

                        key_to_date = key.to_date

                        not_in_time_days.push key if key_to_date < params[:from].to_date || key_to_date > params[:to].to_date

                        # Day of today and future are not allowed
                        if key_to_date >= Date.today 
                            ignore_days.push key
                            next
                        end

                        # Day with already entered data is not allowed to be filled up with planned data
                        ignore_days.push(key) if !value.empty?
                    end

                    employee_data.each do |key, value| 
                        next if ignore_days.include? key

                        if value.nil? || value.empty?
                            value = []

                            next if !daily_data[key] ||
                                daily_data[key].empty? || 
                                daily_data[key][Constants::PLANSTATUS_KEY] < Constants::PS_Vorplan
                        
                            daily_data_object = daily_data[key]
                            teldet_list = []
                            last_teldet = nil
                            last_type = nil

                            if daily_data_object[Constants::TEL_LIST_CLIENT_KEY] && !daily_data_object[Constants::TEL_LIST_CLIENT_KEY].empty?
                                daily_data_object[Constants::TEL_LIST_CLIENT_KEY].each do |tel|
                                    if tel && tel[Constants::TEL_DET_LIST_KEY] && tel[Constants::TEL_DET_LIST_KEY][Constants::ITEMS_KEY] && 
                                        tel[Constants::TEL_DET_LIST_KEY][Constants::ITEMS_KEY].compact.length
                                        teldets = tel[Constants::TEL_DET_LIST_KEY][Constants::ITEMS_KEY].compact

                                        teldets.each do |teldet|
                                            planarea_id = teldet[Constants::PLANAREA_ID_KEY]
                                            planarea = nil
                                            planarea = planareas[planarea_id] if planarea_id > -1
                                            next if planarea.nil?

                                            type = nil
                                            isBreak = false
                                            isFulldayState = false
                                            if teldet[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_SERVICE
                                                type = Constants::TYPE_SERVICE
                                                last_teldet = teldet
                                                last_type = type
                                            elsif teldet[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STANDBY
                                                type = Constants::TYPE_STANDBY
                                                last_teldet = teldet
                                                last_type = type
                                            elsif teldet[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STATE && 
                                                teldet[Constants::TAGESELEMENT_ART_KEY] == Constants::STATE_HOURLY
                                                # Only allow STATE_HOURLY for setting time for
                                                type = Constants::TYPE_STATE
                                                last_teldet = teldet
                                                last_type = type
=begin
                                            elsif teldet[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STATE &&
                                                teldet[Constants::TAGESELEMENT_ART_KEY] == Constants::STATE_FULL
                                                # Mark full day states
                                                type = Constants::TYPE_STATE
                                                last_teldet = teldet
                                                last_type = type
                                                isFulldayState = true
=end
                                            elsif teldet[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_BREAK
                                                von =  teldet[Constants::VON_KEY]
                                                bis =  teldet[Constants::BIS_KEY]
                                                type = last_type
                                                teldet = last_teldet
                                                teldet[Constants::VON_KEY] = von
                                                teldet[Constants::BIS_KEY] = bis
                                                isBreak = true
                                                last_teldet = nil
                                                last_type = nil
                                            end

                                            next if type.nil? || teldet[Constants::REFERENZ_TAGESDATEN_KEY].nil?
                                            
                                            value.push({ 
                                                "planarea": {
                                                    "id": teldet[Constants::PLANAREA_ID_KEY],
                                                    "description": planarea
                                                },
                                                "duty": {
                                                    "id": teldet[Constants::ID_KUERZEL_IST_KEY],
                                                    "short": teldet[Constants::KUERZEL_IST_KEY],
                                                    "long": "#{teldet[Constants::REFERENZ_TAGESDATEN_KEY][Constants::BEZEICHNUNG_KEY]} (#{planarea})",
                                                    "type": type,
                                                    "isBreak": isBreak
                                                },
                                                "from": teldet[Constants::VON_KEY],
                                                "to": teldet[Constants::BIS_KEY],
                                                "status": Constants::ZA_Unbearbeitet,
                                                "isFulldayState": isFulldayState,
                                                "isPlanData": true
                                            })
                                        end
                                    end
                                end
                            end

                            employee_data[key] = value if !value.empty?
                        end
                    end
                end

                render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, employee_data.except!(*not_in_time_days).sort.to_h)
                return
            end    
        end

        # Freigeben
        def update_timesheet
            invalid_key = check_params([:fct_code], [])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?
            
            return render bad_params_for_render(:timesheet) if params[:timesheet].nil?

            # TODO: US-715 - Fake data storage. Replace with storage inside the database when ready.
            # TODO: TK-1212/TK-1214: Adjust Fake call!!
            fake_timesheet = true
            if fake_timesheet
                timesheet_hash = params.permit(:timesheet => {}).to_hash
                timesheet_data = timesheet_hash[Constants::TIMESHEET_KEY]

                employee_id = timesheet_data['employee_id']
                if employee_id.nil? || employee_id.blank?
                    render bad_params_for_render(:employee_id)
                    return
                end
                timesheet_data.except!('employee_id')

                from = timesheet_data['from'] 
                to = timesheet_data['to']
                if from.nil? || from.blank?
                    render bad_params_for_render(:from)
                    return
                elsif to.nil? || to.blank?
                    render bad_params_for_render(:to)
                    return
                end
                timesheet_data.except!('from')
                timesheet_data.except!('to')

                from = Date.parse(from)
                to = Date.parse(to)

                # All duties that don't have a status yet, need to be set as "In Approval".
                new_timesheet_data =
                    timesheet_data.each_with_object({}) do |(day_key, duty_array), new_hash|

                        # puts 'day_key'
                        day = Date.parse(day_key)
                        next if day < from || day > to

                        duty_array_with_status = duty_array.map do |duty|
                            duty[Constants::STATUS_KEY] = Constants::ZA_Freigegeben
                            
                            duty
                        end

                        new_hash[day_key] = duty_array_with_status
                    end

                # Replace data in the current timesheet with the new data.
                employee_file_path = File.dirname(__FILE__)+"/my_pep_timesheet_#{employee_id.to_i}.json"
                employee_file_data = JSON.parse(File.read(employee_file_path))

                new_timesheet_data.keys.each do |day_key|
                    employee_file_data[day_key] = new_timesheet_data[day_key]
                end
                

                File.open(employee_file_path, "w") do |f|
                    f.write(employee_file_data.to_json)
                end

                render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, new_timesheet_data)
            end
        end

        def update_timesheet_item
            invalid_key = check_params([:fct_code], [])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?
            
            return render bad_params_for_render(:timesheet_item) if params[:timesheet_item].nil?

            # TODO: US-715 - Fake data storage. Replace with storage inside the database when ready.
            fake_timesheet = true
            if fake_timesheet
                timesheet_hash = params.permit(:timesheet_item => {}).to_hash
                timesheet_item_data = timesheet_hash[Constants::TIMESHEET_ITEM_KEY]

                employee_id = timesheet_item_data['employee_id']
                if employee_id.nil? || employee_id.blank?
                    render bad_params_for_render(:employee_id)
                    return
                end
                timesheet_item_data.except!('employee_id')

                operation = timesheet_item_data['operation']
                if operation.nil? || ![Constants::TIMESHEET_OPERATION_NEW, Constants::TIMESHEET_OPERATION_UPDATE, Constants::TIMESHEET_OPERATION_DELETE].include?(operation)
                    render bad_params_for_render(:operation)
                    return
                end
                timesheet_item_data.except!('operation')

                # Load data of the current timesheet.
                employee_file_path = File.dirname(__FILE__)+"/my_pep_timesheet_#{employee_id.to_i}.json"
                employee_file_data = JSON.parse(File.read(employee_file_path))
                
                date = Date.parse(timesheet_item_data['date']).strftime(Constants::DATE_FORMAT)
                db_id = timesheet_item_data[Constants::TIMESHEET_DB_ID]

                if employee_file_data[date].nil? || employee_file_data[date].empty?
                    employee_file_data[date] = []
                end
                day_data = []

                if operation == Constants::TIMESHEET_OPERATION_NEW
                    db_id = get_fake_ID_for_entry
                    timesheet_item_data[Constants::TIMESHEET_DB_ID] = db_id
                    timesheet_item_data[Constants::STATUS_KEY] = Constants::ZA_Erfasst_Neu

                    day_data = employee_file_data[date]
                    day_data.push timesheet_item_data
                elsif operation == Constants::TIMESHEET_OPERATION_UPDATE || 
                        operation == Constants::TIMESHEET_OPERATION_DELETE || 
                        employee_file_data[date].nil? || employee_file_data[date].empty?
                    return render bad_params_for_render(:db_id) if db_id.nil? || db_id < 0
                    
                    employee_file_data[date].each do |entry|
                        if operation == Constants::TIMESHEET_OPERATION_UPDATE 
                            if entry[Constants::TIMESHEET_DB_ID] == db_id
                                timesheet_item_data[Constants::STATUS_KEY] = Constants::ZA_Erfasst_Bearbeitet
                                day_data.push timesheet_item_data
                            else
                                entry[Constants::STATUS_KEY] = Constants::ZA_Erfasst_Bearbeitet
                                day_data.push entry
                            end
                        elsif operation == Constants::TIMESHEET_OPERATION_DELETE && entry[Constants::TIMESHEET_DB_ID] != db_id
                            entry[Constants::STATUS_KEY] = Constants::ZA_Erfasst_Bearbeitet
                            day_data.push entry
                        end
                    end

                    db_id = -1 if operation == Constants::TIMESHEET_OPERATION_DELETE
                else
                    render_plan_error(Constants::FROM_TIMESHEET_ITEM, Constants::INVALID_DATA)
                    return
                end

                employee_file_data[date] = day_data
                          
                # Save data of the current timesheet.
                File.open(employee_file_path, "w") do |f|
                    f.write(employee_file_data.to_json)
                end

                render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, db_id)
            end
        end

        def get_fake_ID_for_entry
            fake_file_path = File.dirname(__FILE__)+"/my_pep_fake_timesheet.index"
            if !File.file?(fake_file_path)
                # Initialize index at 0.
                File.open(fake_file_path, 'w') { |file| file.write("0") }

                0
            else
                index = File.read(fake_file_path).to_i
                next_index = index + 1

                File.open(fake_file_path, 'w') { |file| file.write("#{next_index}") }
                next_index
            end
        end

        def load_requests
            invalid_key = check_params([:fct_code, :type], [:employee_id])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?
            
            # TODO: US-761 - Fake data loading. Replace with loading from the database when ready.
            fake_requests = true
            if fake_requests
                requests = load_requests_via_fake_file params[:employee_id].to_i

                requests.keep_if do |request|
                    request[Constants::TYPE_KEY] == params[:type] || params[:type] == Constants::REQUEST_TYPE_ALL
                end

                render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, requests)
                return
            end    
        end

        def create_request
            invalid_key = check_params([:fct_code], [])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?
            return render bad_params_for_render(:request) if params[:request].nil?

            # TODO: US-761 - Fake data storage. Replace with storage inside the database when ready.
            fake_requests = true
            if fake_requests
                request_hash = params.permit(:request => {}).to_hash
                request_data = request_hash["request"]

                employee_id = request_data['employee_id']
                if employee_id.nil? || employee_id.blank?
                    render bad_params_for_render(:employee_id)
                    return
                end
                request_data.except!('employee_id')

                # New requests get the workflow status "new".
                request_data[Constants::STATUS_KEY] = Constants::REQUEST_STATUS_NEW

                # New requests are created now.
                request_data[Constants::CREATED_AT_KEY] = Time.now.iso8601

                # New requests get their index now.
                request_data[Constants::ID_CLIENT_KEY] = next_request_index_via_fake_file()

                requests = load_requests_via_fake_file employee_id.to_i
                requests.push request_data

                # Replace data in the current request file with the new data.
                employee_file_path = File.dirname(__FILE__)+"/my_pep_requests_#{employee_id.to_i}.json"
                
                File.open(employee_file_path, "w") do |f|
                    f.write(requests.to_json)
                end

                render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, true)
            end
        end

        def cancel_request
            invalid_key = check_params([:fct_code], [:employee_id, Constants::REQUEST_ID_KEY])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?

            employee_id = params[:employee_id]

            # TODO: US-761 - Fake data storage. Replace with storage inside the database when ready.
            fake_requests = true
            if fake_requests
                requests = load_requests_via_fake_file employee_id.to_i

                requests.each do |request|
                    request[Constants::STATUS_KEY] = Constants::REQUEST_STATUS_CANCELED if request[Constants::ID_CLIENT_KEY] == params[Constants::REQUEST_ID_KEY].to_i
                end

                # Replace data in the current request file with the new data.
                employee_file_path = File.dirname(__FILE__)+"/my_pep_requests_#{employee_id.to_i}.json"
                
                File.open(employee_file_path, "w") do |f|
                    f.write(requests.to_json)
                end

                render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, true)
            end            
        end

        def get_longterm_account
            invalid_key = check_params([:fct_code, :date], [:employee_id])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?
      
            begin
                response_string, string_pointer = ServiceLayerDll.get_staff_scheduling_longterm_account(
                    current_user.as_json,
                    params[:fct_code],
                    params[:employee_id].to_i,
                    params[:date]
                )
                ServiceLayerDll.free_string(string_pointer)
        
                longterm_account_data, state = extract_data_and_state_from_response_string(response_string)
                
                if state != Rack::Utils.status_code(:ok) || longterm_account_data.nil?
                    render parse_response_string_for_render(response_string)
                    return
                end
        
                render_server_longterm_account_data(longterm_account_data)
            rescue Exception => e
                render parse_exception_for_render(e)
            end
          end

        def get_vacation_account
            invalid_key = check_params([:fct_code], [:planarea_id, :employee_id, :year])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?
      
            begin
                response_string, string_pointer = ServiceLayerDll.get_staff_scheduling_vacation_account(
                    current_user.as_json,
                    params[:fct_code],
                    params[:planarea_id].to_i,
                    params[:employee_id].to_i,
                    params[:year].to_i
                )
                ServiceLayerDll.free_string(string_pointer)
        
                vacation_data, state = extract_data_and_state_from_response_string(response_string)
                
                if state != Rack::Utils.status_code(:ok) || vacation_data.nil?
                    render parse_response_string_for_render(response_string)
                    return
                end
        
                render_server_vacation_data(vacation_data)
            rescue Exception => e
                render parse_exception_for_render(e)
            end
        end

        def get_annual_working_time_account
            invalid_key = check_params([:fct_code], [:employee_id, :year])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?
      
            begin
                response_string, string_pointer = ServiceLayerDll.get_staff_scheduling_annual_working_time_account(
                    current_user.as_json,
                    params[:fct_code],
                    params[:employee_id].to_i,
                    params[:year].to_i
                )
                ServiceLayerDll.free_string(string_pointer)
        
                annual_working_time_account_data, state = extract_data_and_state_from_response_string(response_string)
                
                if state != Rack::Utils.status_code(:ok) || annual_working_time_account_data.nil?
                    render parse_response_string_for_render(response_string)
                    return
                end
        
                render_server_annual_working_time_account_data(annual_working_time_account_data)
            rescue Exception => e
                render parse_exception_for_render(e)
            end
        end

        def get_plan_history
            invalid_key = check_params([:fct_code, :from, :to, :filter], [:planarea_id, :employee_id])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?

            begin
                response_string, string_pointer = ServiceLayerDll.get_staff_scheduling_plan_history(
                    current_user.as_json,
                    params[:fct_code],
                    params[:planarea_id].to_i,
                    params[:employee_id].to_i,
                    params[:from],
                    params[:to],
                    params[:filter].to_s == 'true'
                )
                ServiceLayerDll.free_string(string_pointer)
        
                plan_history_data, state = extract_data_and_state_from_response_string(response_string)
                
                if state != Rack::Utils.status_code(:ok) || plan_history_data.nil?
                    render parse_response_string_for_render(response_string)
                    return
                end
        
                render_server_plan_history_data(plan_history_data)
            rescue Exception => e
                render parse_exception_for_render(e)
            end
        end

        def get_time_bookings
            invalid_key = check_params([:fct_code, :date], [:planarea_id, :employee_id])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?

            # TODO: Replace fake data with server call once it's implemented.
            local_date = DateTime.now

            request_date = DateTime.strptime(params[:date], Constants::DATE_FORMAT)

            # Make sure the local time zone is used.
            request_date = DateTime.new(request_date.year, request_date.month, request_date.day, 0, 0, 0, local_date.zone)
            request_date_next = request_date + 1.day

            is_date_checked = request_date.day % 2 != 0
            is_date_next_checked = request_date_next.day % 2 != 0

            fake_data = [
                {
                    Constants::DATE_KEY => request_date.strftime(Constants::DATE_FORMAT),
                    Constants::TIME_BOOKING_IS_DAY_CHECKED_KEY => is_date_checked,
                    Constants::TIME_BOOKING_KEY => create_fake_time_booking_data_for_date(request_date),
                    Constants::DUTY_KEY => create_fake_duty_booking_data_for_date(request_date),
                    Constants::TIME_BOOKING_BREAKS_KEY => create_fake_breaks_booking_data_for_date(request_date)
                },

                {
                    Constants::DATE_KEY => request_date_next.strftime(Constants::DATE_FORMAT),
                    Constants::TIME_BOOKING_IS_DAY_CHECKED_KEY => is_date_next_checked,
                    Constants::TIME_BOOKING_KEY => create_fake_time_booking_data_for_date(request_date_next),
                    Constants::DUTY_KEY => create_fake_duty_booking_data_for_date(request_date_next),
                    Constants::TIME_BOOKING_BREAKS_KEY => create_fake_breaks_booking_data_for_date(request_date_next)
                }
            ]

            render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, fake_data)
        end

        # Gets the employee right of the current_user
        def get_employee_rights
            invalid_key = check_params([:fct_code], [])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?
      
            begin
              response_string, string_pointer = ServiceLayerDll.get_employee_rights(
                current_user.as_json,
                params[:fct_code]
              )
              ServiceLayerDll.free_string(string_pointer)
      
              employee_rights, state = extract_data_and_state_from_response_string(response_string)
      
              if state != Rack::Utils.status_code(:ok) || employee_rights.nil?
                  return render parse_response_string_for_render(response_string)
              end
      
              # Extract the necessary values for the rights.
              user_config = employee_rights[Constants::RIGHTS_USER_CONFIG_KEY]
              function_rights = employee_rights[Constants::RIGHTS_FUNCTION_RIGHTS_KEY]
      
              max_day_change_duty = user_config[Constants::RIGHTS_MAX_DAY_CHANGE_DUTY_KEY]
              max_day_change_standby = user_config[Constants::RIGHTS_MAX_DAY_CHANGE_STANDBY_KEY]
              max_day_change_timesheet = user_config[Constants::RIGHTS_MAX_DAY_CHANGE_TIMESHEET_KEY]
      
              workflow_vacation = user_config[Constants::RIGHTS_WORKFLOW_VACATION_KEY]
              workflow_booking_correction = user_config[Constants::RIGHTS_WORKFLOW_BOOKING_CORRECTION_KEY]
              workflow_overtime_reduction = user_config[Constants::RIGHTS_WORKFLOW_OVERTIME_REDUCTION_KEY]
              workflow_ill_without_disabled = user_config[Constants::RIGHTS_WORKFLOW_ILL_WITHOUT_DISABLEDKEY]
      
              rights_hash = {
                  Constants::RIGHTS_MAX_DAY_CHANGE_DUTY_KEY => max_day_change_duty,
                  Constants::RIGHTS_MAX_DAY_CHANGE_STANDBY_KEY => max_day_change_standby,
                  Constants::RIGHTS_MAX_DAY_CHANGE_TIMESHEET_KEY => max_day_change_timesheet,
                  Constants::RIGHTS_WORKFLOW_VACATION_KEY => workflow_vacation,
                  Constants::RIGHTS_WORKFLOW_BOOKING_CORRECTION_KEY => workflow_booking_correction,
                  Constants::RIGHTS_WORKFLOW_OVERTIME_REDUCTION_KEY => workflow_overtime_reduction,
                  Constants::RIGHTS_WORKFLOW_ILL_WITHOUT_DISABLEDKEY => workflow_ill_without_disabled,
                  Constants::RIGHTS_START_KEY => false,
                  Constants::RIGHTS_DUTY_INPUT_KEY => false,
                  Constants::RIGHTS_DUTY_WISH_KEY => false,
                  Constants::RIGHTS_STANDBY_INPUT_KEY => false,
                  Constants::RIGHTS_STANDBY_WISH_KEY => false,
                  Constants::RIGHTS_PLANAREA_VISIBLE_KEY => false,
                  Constants::RIGHTS_CHANGE_APPROVED_DUTY_TOKENS_KEY => false,
                  Constants::RIGHTS_CHANGE_APPROVED_STANDBY_TOKENS_KEY => false,
                  Constants::RIGHTS_WARNINGS_VISIBLE_KEY => false,
                  Constants::RIGHTS_ONLY_APPROVED_DUTY_PLANS_VISIBLE_KEY => false,
                  Constants::RIGHTS_TIMESHEET_KEY => false,
                  Constants::RIGHTS_DAY_INSPECTOR_KEY => false,
                  Constants::RIGHTS_CUMULATION_KEY => false,
                  Constants::RIGHTS_FUTURE_DISABLED_KEY => false
              }
      
              rights_key_enum_array = [
                  [Constants::RIGHTS_START, Constants::RIGHTS_START_KEY],
                  [Constants::RIGHTS_DUTY_INPUT, Constants::RIGHTS_DUTY_INPUT_KEY],
                  [Constants::RIGHTS_DUTY_WISH, Constants::RIGHTS_DUTY_WISH_KEY],
                  [Constants::RIGHTS_STANDBY_INPUT, Constants::RIGHTS_STANDBY_INPUT_KEY],
                  [Constants::RIGHTS_STANDBY_WISH, Constants::RIGHTS_STANDBY_WISH_KEY],
                  [Constants::RIGHTS_PLANAREA_VISIBLE, Constants::RIGHTS_PLANAREA_VISIBLE_KEY],
                  [Constants::RIGHTS_CHANGE_APPROVED_DUTY_TOKENS, Constants::RIGHTS_CHANGE_APPROVED_DUTY_TOKENS_KEY],
                  [Constants::RIGHTS_CHANGE_APPROVED_STANDBY_TOKENS, Constants::RIGHTS_CHANGE_APPROVED_STANDBY_TOKENS_KEY],
                  [Constants::RIGHTS_WARNINGS_VISIBLE, Constants::RIGHTS_WARNINGS_VISIBLE_KEY],
                  [Constants::RIGHTS_ONLY_APPROVED_DUTY_PLANS_VISIBLE, Constants::RIGHTS_ONLY_APPROVED_DUTY_PLANS_VISIBLE_KEY],
                  [Constants::RIGHTS_TIMESHEET, Constants::RIGHTS_TIMESHEET_KEY],
                  [Constants::RIGHTS_DAY_INSPECTOR, Constants::RIGHTS_DAY_INSPECTOR_KEY],
                  [Constants::RIGHTS_CUMULATION, Constants::RIGHTS_CUMULATION_KEY],
                  [Constants::RIGHTS_FUTURE_DISABLED, Constants::RIGHTS_FUTURE_DISABLED_KEY]
              ]
      
              function_rights_array = function_rights[Constants::ITEMS_KEY]
      
              function_rights_array.each do |rights_array|
                  rights_key_enum_array.each do |rights_enum, rights_key|
                      if rights_array.include? rights_enum
                          rights_hash[rights_key] = true
                      end
                  end
              end
      
              render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, rights_hash)
            rescue Exception => e
              render parse_exception_for_render(e)
            end
          end

        def get_definitions_for_timesheet
            invalid_key = check_params([:fct_code, :date], [:employee_id])
            return render bad_params_for_render(invalid_key) unless invalid_key.blank?
            
            begin
                employee_id = params[:employee_id]
                response_string, string_pointer = ServiceLayerDll.get_definitions_for_timesheet(
                    current_user.as_json,
                    params[:fct_code],
                    employee_id.to_i,
                    params[:date]
                )
                ServiceLayerDll.free_string(string_pointer)
        
                definitions_for_timesheet, state = extract_data_and_state_from_response_string(response_string)
                
                if state != Rack::Utils.status_code(:ok) || definitions_for_timesheet.nil?
                    render parse_response_string_for_render(response_string)
                    return
                end
                
                [Constants::MIT_LIST_KEY, Constants::PLANBEREICHE_KEY].each do |key| 
                    if definitions_for_timesheet[key].nil? 
                        render_plan_error(Constants::FROM_DEFINITIONS_OF_TIMESHEET, Constants::INVALID_DATA)
                        return
                    end
                end
                
                mit_list = definitions_for_timesheet[Constants::MIT_LIST_KEY]
                
                # Determine Planbereiche
                planbereiche = {}
                if definitions_for_timesheet[Constants::PLANBEREICHE_KEY].present? && definitions_for_timesheet[Constants::PLANBEREICHE_KEY][Constants::ITEMS_KEY].present?
                    definitions_for_timesheet[Constants::PLANBEREICHE_KEY][Constants::ITEMS_KEY].compact.each { |planbereich| planbereiche[planbereich[Constants::ID_STATION_KEY]] = planbereich[Constants::PLANBEREICH_KEY] }
                end

                # Determine Mitarbeiterliste
                employeeArray = mit_list[Constants::ITEMS_KEY].compact
                if employeeArray.blank?
                    render_plan_error(Constants::FROM_DEFINITIONS_OF_TIMESHEET, Constants::EMPTY_EMPLOYEES)
                    return
                end

                # Determine Mitarbeiter and TKS-Array
                employee = nil
                tksArray = nil
                for employee in employeeArray
                    if employee[Constants::ID_SERVER_KEY] == employee_id.to_i
                        tksArray = employee[Constants::TKS_LIST_KEY]
                        break
                    end
                end

                # Check determined values from above
                if employee.nil? || tksArray.nil?
                    render_plan_error(Constants::FROM_DEFINITIONS_OF_TIMESHEET, Constants::EMPTY_DAILY_DATA)
                    return
                end
                tksItemArray = tksArray[Constants::ITEMS_KEY]
                if tksItemArray.nil?
                    render_plan_error(Constants::FROM_DEFINITIONS_OF_TIMESHEET, Constants::EMPTY_DAILY_DATA)
                    return
                end

                # Filter out null values
                tksItemArray = tksItemArray.compact
                
                dienste_def_list = []
                bereit_def_list = []
                status_def_list = []

                # Iterate through the TKS-Array
                for tksItem in tksItemArray
                    next if tksItem.nil? || tksItem[Constants::DATUM_KEY].nil?
                    begin
                        tksItemDate = tksItem[Constants::DATUM_KEY].to_time
                    rescue
                        next
                    end

                    next if tksItemDate.strftime(Constants::DATE_FORMAT) != params[:date]

                    next if tksItem[Constants::TEL_LIST_SERVER_KEY].nil? || tksItem[Constants::TEL_LIST_SERVER_KEY][Constants::ITEMS_KEY].nil?
                    tel_items = tksItem[Constants::TEL_LIST_SERVER_KEY][Constants::ITEMS_KEY].compact
                    next if !tel_items.length
                
                    for tel_item in tel_items
                        next if tel_item.nil? || tel_item[Constants::DATUM_KEY].nil? || tel_item[Constants::TAGESELEMENT_KATEGORIE_KEY].nil?

                        #TODO: Could also check anzahl property from the tel_item! (anzahl_TELDET)
                        next if tel_item[Constants::TEL_DET_LIST_KEY].nil? || tel_item[Constants::TEL_DET_LIST_KEY][Constants::ITEMS_KEY].nil?
                        teldet_items = tel_item[Constants::TEL_DET_LIST_KEY][Constants::ITEMS_KEY].compact
                        next if teldet_items.length == 0

                        telItem = check_state_type(tel_item)

                        for telDetItem in teldet_items
                            next if telDetItem.nil? || telDetItem[Constants::DATUM_KEY].nil? || 
                                    telDetItem[Constants::TAGESELEMENT_KATEGORIE_KEY].nil? ||
                                    telDetItem[Constants::REFERENZ_TAGESDATEN_KEY].nil?

                            refItem = telDetItem[Constants::REFERENZ_TAGESDATEN_KEY]

                            refItem[Constants::PLANBEREICH_KEY] = ""
                            if planbereiche[refItem[Constants::ID_STATION_KEY]].present?
                                refItem[Constants::PLANBEREICH_KEY] = planbereiche[refItem[Constants::ID_STATION_KEY]]
                                refItem[Constants::BEZEICHNUNG_KEY] = refItem[Constants::BEZEICHNUNG_KEY] + " (#{refItem[Constants::PLANBEREICH_KEY]})"
                            end

                            # Sort TEL-Element by his category and push it into the right Array
                            if telDetItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_SERVICE
                                dienste_def_list.push(refItem) unless dienste_def_list.include? refItem 
                            elsif telDetItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STANDBY #|| telDetItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STANDBY_OPERATION
                                bereit_def_list.push(refItem) unless bereit_def_list.include? refItem
                            elsif telDetItem[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STATE && telDetItem[Constants::STATE_TYPE_KEY] == Constants::STATE_HOURLY
                                status_def_list.push(refItem) unless status_def_list.include? refItem
                            end
                        end
                    end
                end

                data = {
                    Constants::DIENSTE_DEF_LIST_KEY => dienste_def_list.each {|def_item| def_item[:type] = Constants::TYPE_SERVICE},
                    Constants::BEREIT_DEF_LIST_KEY => bereit_def_list.each {|def_item| def_item[:type] = Constants::TYPE_STANDBY},
                    Constants::STATUS_DEF_LIST_KEY => status_def_list.each {|def_item| def_item[:type] = Constants::TYPE_STATE}
                }
        
                render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, data)
            rescue Exception => e
                render parse_exception_for_render(e)
            end
        end

        private

        def handle_get_definitions_response_string(response_string)
            time_marker = start_timer
            definitions, state = extract_data_and_state_from_response_string(response_string)
            stop_timer(time_marker, "Render making object from hash")

            time_marker = start_timer
            if state != Rack::Utils.status_code(:ok) || definitions.nil?
                render parse_response_string_for_render(response_string)
                return
            end

            [Constants::DIENSTE_DEF_LIST_KEY, Constants::BEREIT_DEF_LIST_KEY, Constants::STATUS_DEF_LIST_KEY].each do |key| 
                if definitions[key].nil? 
                    render_plan_error(Constants::FROM_DEFINITIONS, Constants::INVALID_DATA)
                    return
                end
            end

            data = {
                Constants::DIENSTE_DEF_LIST_KEY => definitions[Constants::DIENSTE_DEF_LIST_KEY][Constants::ITEMS_KEY].compact.each {|def_item| def_item[:type] = Constants::TYPE_SERVICE},
                Constants::BEREIT_DEF_LIST_KEY => definitions[Constants::BEREIT_DEF_LIST_KEY][Constants::ITEMS_KEY].compact.each {|def_item| def_item[:type] = Constants::TYPE_STANDBY},
                Constants::STATUS_DEF_LIST_KEY => definitions[Constants::STATUS_DEF_LIST_KEY][Constants::ITEMS_KEY].compact.each {|def_item| def_item[:type] = Constants::TYPE_STATE}
            }

            stop_timer(time_marker, "Render iterating & sorting")

            render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, data)
        end

        def create_empty_daily_data_object(tksItem)
            dailyDataObject = {}

            dailyDataObject[Constants::ZEITSTEMPEL_CLIENT_KEY] = tksItem[Constants::ZEITSTEMPEL_SERVER_KEY]

            dailyDataObject[Constants::SERVICES_KEY] = []
            dailyDataObject[Constants::STANDBYS_KEY] = []
            dailyDataObject[Constants::STATES_KEY] = []
            dailyDataObject[Constants::COLOR_INFOS_KEY] = []

            dailyDataObject[Constants::PLANSTATUS_KEY] = Constants::PS_Fixiert
            dailyDataObject[Constants::FERIEN_KEY] = {}

            dailyDataObject[Constants::DAY_FROM_KEY] = ""
            dailyDataObject[Constants::DAY_TO_KEY] = ""

            dailyDataObject[Constants::FULL_DAY_STATE_KEY] = Constants::STATE_FREE
            dailyDataObject[Constants::OVERNIGHT_KEY] = false

            dailyDataObject[Constants::PLANAREAS_KEY] = []

            dailyDataObject[Constants::TEL_LIST_CLIENT_KEY] = []
            
            # Set the HolidayDay-Infos
            dailyDataObject[Constants::IS_HOLIDAY_DAY_KEY] = tksItem[Constants::IS_FEIERTAG_KEY]
            dailyDataObject[Constants::HOLIDAY_DAY_REF_KEY] = tksItem[Constants::FEIERTAG_DEF_KEY]

            dailyDataObject
        end

        # FAKE
        def load_requests_via_fake_file(employee_id)
            fake_file_path = File.dirname(__FILE__)+"/my_pep_fake_requests.json"
            employee_file_path = File.dirname(__FILE__)+"/my_pep_requests_#{employee_id}.json"
            if !File.file?(employee_file_path)
                FileUtils.cp(fake_file_path, employee_file_path)
            end
            JSON.parse(File.read(employee_file_path))
        end

        # FAKE
        def next_request_index_via_fake_file
            fake_file_path = File.dirname(__FILE__)+"/my_pep_fake_requests.index"
            if !File.file?(fake_file_path)
                # Initialize index at 0.
                File.open(fake_file_path, 'w') { |file| file.write("0") }

                0
            else
                index = File.read(fake_file_path).to_i
                next_index = index + 1

                File.open(fake_file_path, 'w') { |file| file.write("#{next_index}") }
                next_index
            end
        end

        # FAKE
        def add_warnings(tksItemDate, dailyDataObject)
            day = tksItemDate.strftime("%d")
            if day == '03' || day == '10' || day == '17' || day == '22' || day == '30'
                dailyDataObject[Constants::WARNING_LEFT_KEY].push "Stundenanzahl überschritten!"
            end
            if day == '02' || day == '14' || day == '26' || day == '31'
                dailyDataObject[Constants::WARNING_RIGHT_KEY].push "Arbeitszeitverletzung!"
            end
        end

        def render_plan_error(plan, reason)
            render plan_error(plan, reason)
        end

        def plan_error(plan, reason)
            return create_json_error_response("#{Constants::ERROR_TEXT} #{plan} - #{reason}")
        end

        def check_state_type(tel_item)
            tel_item[Constants::STATUS_HOUR_KEY] = []
            
            #puts "statusArt == #{tel_item[Constants::TAGESELEMENT_ART_KEY]} " if (tel_item[Constants::TAGESELEMENT_KATEGORIE_KEY] == Constants::TAGESELEMENT_KATEGORIE_STATE)

            if (tel_item[Constants::TAGESELEMENT_KATEGORIE_KEY] != 5)
                tel_item[Constants::STATE_TYPE_KEY] = "-"
            elsif tel_item[Constants::TAGESELEMENT_ART_KEY] == Constants::STATE_FULL                   #0
                tel_item[Constants::STATE_TYPE_KEY] = Constants::STATE_FULL
            elsif tel_item[Constants::TAGESELEMENT_ART_KEY] == Constants::STATE_HOURLY_OVERLAPPING     #1
                tel_item[Constants::STATE_TYPE_KEY] = Constants::STATE_HOURLY_OVERLAPPING
            elsif tel_item[Constants::TAGESELEMENT_ART_KEY] == Constants::STATE_HOURLY                 #2
                tel_item[Constants::STATE_TYPE_KEY] = Constants::STATE_HOURLY
            end
            
            tel_item
        end

        def check_print_kuerzel(tel_or_teldet_item, employee_id)
            print_kuerzel = ""

            if current_user.pepMitarbeiterID != employee_id && tel_or_teldet_item && tel_or_teldet_item[Constants::REFERENZ_TAGESDATEN_KEY] && 
                tel_or_teldet_item[Constants::REFERENZ_TAGESDATEN_KEY][Constants::PRINT_KUERZEL_KEY] && tel_or_teldet_item[Constants::REFERENZ_TAGESDATEN_KEY][Constants::PRINT_KUERZEL_KEY].length 
                print_kuerzel = tel_or_teldet_item[Constants::REFERENZ_TAGESDATEN_KEY][Constants::PRINT_KUERZEL_KEY]
            end    
            
            print_kuerzel
        end

        #TODO: Could also check anzahl property from the tel_item always! (anzahl_TELDET)
        def check_for_data_in_tks_item(tksItem, optimized_call = false)
            valid_data = false

            if tksItem[Constants::TEL_LIST_SERVER_KEY].present? && tksItem[Constants::TEL_LIST_SERVER_KEY][Constants::ITEMS_KEY].present?
                tel_items = tksItem[Constants::TEL_LIST_SERVER_KEY][Constants::ITEMS_KEY].compact
                return valid_data if !tel_items.length

                for tel_item in tel_items
                    valid_data = tel_item.present? && tel_item[Constants::DATUM_KEY].present? && tel_item[Constants::TAGESELEMENT_KATEGORIE_KEY].present? &&
                                 ((optimized_call && tel_item[Constants::ANZAHL_TEL_DET_KEY] > 0) || (tel_item[Constants::TEL_DET_LIST_KEY].present? && tel_item[Constants::TEL_DET_LIST_KEY][Constants::ITEMS_KEY].present? &&
                                 tel_item[Constants::TEL_DET_LIST_KEY][Constants::ITEMS_KEY].compact.length > 0))

                    break if valid_data
                end
            end

            valid_data
        end

        def determine_informations(employee, tksItemArray, asked_for_date, current_header, smallest_change)
            # TODO: Think about fulltime state for "rest_time"
            informations = {
                Constants::REST_TIME_KEY => {
                    Constants::TITLE_KEY => Constants::REST_TIME_TITLE,
                    Constants::VALUE_KEY => Constants::FREI,
                    Constants::DETAILS_KEY => "",
                    Constants::TYPE_KEY => Constants::TYPE_THREE_LINES
                }
            }
            begin
                current = Time.now
                work_until = nil

                if (tksItemArray.length > 0)
                    tksItem = tksItemArray[0]
                    
                    tksItemDate = tksItem[Constants::DATUM_KEY].to_date
                    if tksItemDate == current.to_date && tksItem[Constants::TEL_LIST_SERVER_KEY].present? && tksItem[Constants::TEL_LIST_SERVER_KEY][Constants::ITEMS_KEY].present?
                        tel_items = tksItem[Constants::TEL_LIST_SERVER_KEY][Constants::ITEMS_KEY].compact
                        if tel_items.length
                            for telItem in tel_items
                                if telItem.present? && telItem[Constants::DATUM_KEY].present? && telItem[Constants::TAGESELEMENT_KATEGORIE_KEY].present? && telItem[Constants::TEL_DET_LIST_KEY].present? || telItem[Constants::TEL_DET_LIST_KEY][Constants::ITEMS_KEY].present?
                                    teldet_items = telItem[Constants::TEL_DET_LIST_KEY][Constants::ITEMS_KEY].compact
                                    for teldetItem in teldet_items
                                        if teldetItem[Constants::BIS_KEY].present? && !teldetItem[Constants::BIS_KEY].blank?
                                            work_until = teldetItem[Constants::BIS_KEY].to_time 
                                        end
                                    end
                                end
                            end
                        end
                    end
                end

                now = current.to_time
                if work_until.present?
                    if work_until < now
                        informations[Constants::REST_TIME_KEY] = {
                            Constants::TITLE_KEY => Constants::REST_TIME_TITLE,
                            Constants::VALUE_KEY => Time.at((now - work_until).to_i.abs).utc.strftime(Constants::TIME_FORMAT),
                            Constants::DETAILS_KEY => "",
                            Constants::TYPE_KEY => Constants::TYPE_THREE_LINES
                        }
                    else
                        informations[Constants::REST_TIME_KEY] = {
                            Constants::TITLE_KEY => Constants::REST_TIME_TITLE,
                            Constants::VALUE_KEY => Constants::READY,
                            Constants::DETAILS_KEY => "",
                            Constants::TYPE_KEY => Constants::TYPE_THREE_LINES
                        }
                    end
                end
            rescue Exception => e
                # Silently ignore errors
            end

            if smallest_change.present?
                informations[Constants::LAST_CHANGE_KEY] = {
                    Constants::TITLE_KEY => "#{Constants::LAST_CHANGE_ON_TITLE} #{smallest_change.strftime(Constants::DATE_FORMAT)}.",
                    Constants::VALUE_KEY => "",
                    Constants::DETAILS_KEY => "",
                    Constants::TYPE_KEY => Constants::TYPE_TEXT_ONLY
                }
                diff_in_days = ((Time.now.to_time - smallest_change) / 86400).to_i
                since = (diff_in_days).to_s + " #{Constants::DAYS}"
                since = Constants::TODAY if diff_in_days == 0
                since = Constants::YESTERDAY if diff_in_days == 1
                since = Constants::THE_DAY_BEFORE_YESTERDAY if diff_in_days == 2
                informations[Constants::LAST_CHANGE_DURATION_KEY] = {
                    Constants::TITLE_KEY => "#{Constants::SEIT} #{since} #{Constants::NO_LAST_CHANGES_DURATION_TITLE}",
                    Constants::VALUE_KEY => "",
                    Constants::DETAILS_KEY => "",
                    Constants::TYPE_KEY => Constants::TYPE_TEXT_ONLY
                }
            else
                informations[Constants::LAST_CHANGE_DURATION_KEY] = {
                    Constants::TITLE_KEY => Constants::NO_LAST_CHANGE_DURATION_TITLE,
                    Constants::VALUE_KEY => "",
                    Constants::DETAILS_KEY => "",
                    Constants::TYPE_KEY => Constants::TYPE_TEXT_ONLY
                }
            end

            if asked_for_date.present? && employee[Constants::URLAUB_DATALIST_KEY].present? && employee[Constants::URLAUB_DATALIST_KEY][Constants::ITEMS_KEY].present? && employee[Constants::URLAUB_DATALIST_KEY][Constants::ITEMS_KEY].length &&
                for urlaub in employee[Constants::URLAUB_DATALIST_KEY][Constants::ITEMS_KEY].compact
                    if urlaub[Constants::JAHR_KEY].present? && urlaub[Constants::JAHR_KEY] == asked_for_date.year && urlaub[Constants::GENOMMEN_KEY].present? && urlaub[Constants::ANSPRUCH_KEY].present?
                        informations[Constants::HOLIDAYS_KEY] = {
                            Constants::TITLE_KEY => "#{Constants::AVAILABLE_HOLIDAY_DAYS} #{asked_for_date.year}",
                            Constants::VALUE_KEY => "#{urlaub[Constants::GENOMMEN_KEY]}/#{urlaub[Constants::ANSPRUCH_KEY]}",
                            Constants::DETAILS_KEY => "",
                            Constants::TYPE_KEY => Constants::TYPE_BIG_VALUE
                        }
                    end
                end
            end

            soll = 0
            ist = 0
            value = ""
            if current_header[Constants::SOLL_KEY].present? && current_header[Constants::IST_KEY].present?
                soll = current_header[Constants::SOLL_KEY].to_time.seconds_since_midnight
                ist = current_header[Constants::IST_KEY].to_time.seconds_since_midnight
            end
            if ist == 0 && soll != 0
                balance = soll / 3600

                value = hours_to_hours_and_minutes_string(balance)
            elsif ist != 0 && soll != 0
                balance = (soll / ist) / 3600

                value = hours_to_hours_and_minutes_string(balance)
            end
                informations[Constants::BALANCE_WORKTIME_KEY] = {
                    Constants::TITLE_KEY => Constants::BALANCE_WORKTIME_TITLE,
                    Constants::VALUE_KEY => value,
                    Constants::DETAILS_KEY => "",
                    Constants::TYPE_KEY => Constants::TYPE_THREE_LINES
                } unless value.blank?

            informations
        end

        def hours_to_hours_and_minutes_string(hours)
            only_hours = hours.floor
            only_hours_after_decimal_part = only_hours.modulo(1)
            only_minutes = (only_hours_after_decimal_part * 60).round

            "#{only_hours} Std. #{only_minutes} Min."
        end

        def render_server_longterm_account_data(server_data)
            accountData = []
            if server_data[Constants::EINTRAEGE_ZUM_DATUM].present? && !server_data[Constants::EINTRAEGE_ZUM_DATUM][Constants::ITEMS_KEY].nil? 
                for eintrag in server_data[Constants::EINTRAEGE_ZUM_DATUM][Constants::ITEMS_KEY].compact
                    entry = {
                        Constants::COMMENT_KEY => eintrag[Constants::BEMERKUNG_KEY],
                        Constants::CREATED_AT_KEY => eintrag[Constants::LTA_ERSTELLT_AM_KEY],
                        Constants::CREATED_BY_KEY => eintrag[Constants::LTA_ERSTELLT_VON_KEY],
                        Constants::DATE_KEY => eintrag[Constants::DATUM_KEY],
                        Constants::HOURS_KEY => eintrag[Constants::STUNDEN_KEY],
                        Constants::UPDATED_AT_KEY => eintrag[Constants::LTA_GEAENDERT_AM_KEY],
                        Constants::UPDATED_BY_KEY => eintrag[Constants::LTA_GEAENDERT_VON_KEY]
                    }

                    accountData.push entry
                end
            else
                render_plan_error(Constants::FROM_LZK, Constants::INVALID_DATA)
                return
            end

            if server_data[Constants::KONTOSTAND_KEY].nil? || server_data[Constants::END_KONTOSTAND_KEY].nil?
                render_plan_error(Constants::FROM_LZK, Constants::INVALID_DATA)
                return
            end

            data = {
                Constants::ACCOUNT_BALANCE_END_KEY => server_data[Constants::END_KONTOSTAND_KEY],
                Constants::ACCOUNT_BALANCE_TO_DATE_KEY => server_data[Constants::KONTOSTAND_KEY],
                Constants::ACCOUNT_DATA_KEY => accountData
            }

            render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, data)
        end

        def render_server_vacation_data(server_data)
            full_night      = number_from_hash_default(server_data, Constants::VACATION_SERVER_FULL_NIGHT_KEY)
            full_shift      = number_from_hash_default(server_data, Constants::VACATION_SERVER_FULL_SHIFT_KEY)
            full_disability = number_from_hash_default(server_data, Constants::VACATION_SERVER_FULL_DISABILITY_KEY)
            full_other      = number_from_hash_default(server_data, Constants::VACATION_SERVER_FULL_OTHER_KEY)
    
            partial_psychiatry  = number_from_hash_default(server_data, Constants::VACATION_SERVER_PARTIAL_PSYCHIATRY_KEY)
            partial_other_1     = number_from_hash_default(server_data, Constants::VACATION_SERVER_PARTIAL_OTHER_1_KEY)
            partial_other_2     = number_from_hash_default(server_data, Constants::VACATION_SERVER_PARTIAL_OTHER_2_KEY)
            partial_reason      = server_data[Constants::VACATION_SERVER_PARTIAL_REASON_KEY]
    
            whole_year          = number_from_hash_default(server_data, Constants::VACATION_SERVER_WHOLE_YEAR_KEY)
            whole_previous_year = number_from_hash_default(server_data, Constants::VACATION_SERVER_WHOLE_PREVIOUS_YEAR_KEY)
            whole_correction    = number_from_hash_default(server_data, Constants::VACATION_SERVER_WHOLE_CORRECTION_KEY)
            whole_taken         = number_from_hash_default(server_data, Constants::VACATION_SERVER_WHOLE_TAKEN_KEY)
            whole_planned       = number_from_hash_default(server_data, Constants::VACATION_SERVER_WHOLE_PLANNED_KEY)
            whole_approved      = number_from_hash_default(server_data, Constants::VACATION_SERVER_WHOLE_APPROVED_KEY)
            whole_carryover     = number_from_hash_default(server_data, Constants::VACATION_SERVER_WHOLE_CARRYOVER_KEY)
            whole_comment       = server_data[Constants::VACATION_SERVER_WHOLE_COMMENT_KEY]
    
            avz_claim           = number_from_hash_default(server_data, Constants::VACATION_SERVER_AVZ_CLAIM_KEY)
            avz_taken           = number_from_hash_default(server_data, Constants::VACATION_SERVER_AVZ_TAKEN_KEY)

            sum_full = full_night + full_shift + full_disability + full_other
            sum_partial = partial_psychiatry + partial_other_1 + partial_other_2
    
            data = {
                :full => {
                    :night => full_night,
                    :shift => full_shift,
                    :disability => full_disability,
                    :other => full_other,
                    :sum => sum_full
                },
    
                :partial => {
                    :psychiatry => partial_psychiatry,
                    :other_1 => partial_other_1,
                    :other_2 => partial_other_2,
                    :sum => sum_partial,
                    :reason => partial_reason
                },
    
                :whole => {
                    :year => whole_year,
                    :previous_year => whole_previous_year,
                    :extra => sum_full + sum_partial,
                    :correction => whole_correction,
                    :taken => whole_taken,
                    :planned => whole_planned,
                    :approved => whole_approved,
                    :carryover => whole_carryover,
                    :comment => whole_comment
                },
    
                :avz => {
                    :claim => avz_claim,
                    :taken => avz_taken
                }
            }
    
            render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, data)
        end

        def render_server_annual_working_time_account_data(server_data)
            ohne_plandaten = render_server_annual_working_time_account_data_part(server_data[Constants::OHNE_PLANDATEN_KEY])
            return if ohne_plandaten.nil?
            
            mit_plandaten = render_server_annual_working_time_account_data_part(server_data[Constants::MIT_PLANDATEN_KEY])
            return if mit_plandaten.nil?

            data = {
                Constants::WITHOUT_CALCULATION_KEY => mit_plandaten,
                Constants::WITH_CALCULATION_KEY => mit_plandaten
            }

            render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, data)
        end

        def render_server_annual_working_time_account_data_part(server_data_part)
            
            months = {}
            if server_data_part[Constants::MONATE_KEY].present? && server_data_part[Constants::MONATE_KEY][Constants::ITEMS_KEY].present?
                for monat in server_data_part[Constants::MONATE_KEY][Constants::ITEMS_KEY].compact
                    months.merge!(render_server_annual_working_time_account_data_item(monat))  
                end
            end
            if months.blank? || months.length != 12
                render_plan_error(Constants::FROM_JAZK, Constants::INVALID_DATA)
                return nil
            end

            quarters = {}
            if server_data_part[Constants::QUARTALE_KEY].present? && server_data_part[Constants::QUARTALE_KEY][Constants::ITEMS_KEY].present?
                for quartal in server_data_part[Constants::QUARTALE_KEY][Constants::ITEMS_KEY].compact
                    quarters.merge!(render_server_annual_working_time_account_data_item(quartal))  
                end
            end
            if quarters.blank? || quarters.length != 4
                render_plan_error(Constants::FROM_JAZK, Constants::INVALID_DATA)
                return nil
            end

            halfyears = {}
            if server_data_part[Constants::HALBJAHRE_KEY].present? && server_data_part[Constants::HALBJAHRE_KEY][Constants::ITEMS_KEY].present?
                for halbjahr in server_data_part[Constants::HALBJAHRE_KEY][Constants::ITEMS_KEY].compact
                    halfyears.merge!(render_server_annual_working_time_account_data_item(halbjahr))  
                end
            end
            if halfyears.blank? || halfyears.length != 2
                render_plan_error(Constants::FROM_JAZK, Constants::INVALID_DATA)
                return nil
            end

            sumfull = {}
            if server_data_part[Constants::GESAMT_KEY].present? 
                sumfull.merge!(render_server_annual_working_time_account_data_item(server_data_part[Constants::GESAMT_KEY], false))  
            end
            if sumfull.blank?
                render_plan_error(Constants::FROM_JAZK, Constants::INVALID_DATA)
                return nil
            end

            if !server_data_part.has_key?(Constants::PROZENT_KEY) || !server_data_part.has_key?(Constants::DATA_POSITIV_SERVER_KEY)
                render_plan_error(Constants::FROM_JAZK, Constants::INVALID_DATA)
                return nil
            end    

            {
                Constants::MONTH_DATA_KEY => months,
                Constants::QUARTER_DATA_KEY => quarters,
                Constants::HALFYEAR_DATA_KEY => halfyears,
                Constants::SUM_FULL_DATA_KEY => sumfull,
                Constants::ACTUAL_PERCENTAGE_KEY => server_data_part[Constants::PROZENT_KEY],
                Constants::DATA_POSITIV_CLIENT_KEY => server_data_part[Constants::DATA_POSITIV_SERVER_KEY]
            }
        end

        def render_server_annual_working_time_account_data_item(server_data_item, with_key = true)
            hash = 
            {
                Constants::IS_G_KEY => { Constants::VALUE_KEY => server_data_item.has_key?(Constants::GENEHMIGT_KEY) ? server_data_item[Constants::GENEHMIGT_KEY] : false},
                Constants::IS_A_KEY => { Constants::VALUE_KEY => server_data_item.has_key?(Constants::ABGERECHNET_KEY) ? server_data_item[Constants::ABGERECHNET_KEY] : false},
                Constants::TARGET_HOURS_KEY => number_from_hash_default_packed(server_data_item, Constants::SOLL_KEY),
                Constants::ACTUAL_HOURS_KEY => number_from_hash_default_packed(server_data_item, Constants::IST_KEY),
                Constants::STANDBY_KEY => number_from_hash_default_packed(server_data_item, Constants::BEREITSCHAFTEN_KEY),
                Constants::CORRECTION_KEY => number_from_hash_default_packed(server_data_item, Constants::KORREKTUR_KEY),
                Constants::PAID_HOURS_KEY => number_from_hash_default_packed(server_data_item, Constants::STUNDEN_BEZAHLT_KEY),
                Constants::PAID_STANDBY_KEY => number_from_hash_default_packed(server_data_item, Constants::BEREITSCHAFTEN_BEZAHLT_KEY),
                Constants::PAID_VACATION_KEY => number_from_hash_default_packed(server_data_item, Constants::FEIERTAGE_BEZAHLT_KEY),
                Constants::FACTORING_KEY => number_from_hash_default_packed(server_data_item, Constants::GENEHMIGT_KEY),
                Constants::LONG_TERM_ACCOUNT_KEY => number_from_hash_default_packed(server_data_item, Constants::LANGZEITKONTO_KEY),
                Constants::BALANCE_OLD_KEY => number_from_hash_default_packed(server_data_item, Constants::SALDO_ALT_KEY),
                Constants::BALANCE_NEW_KEY => number_from_hash_default_packed(server_data_item, Constants::SALDO_NEU_KEY)
            }

            if with_key
                hash = 
                {
                    "#{server_data_item[Constants::KENNUNG_KEY]}" => hash                    
                }
            end
            
            hash
        end

        def render_server_plan_history_data(server_data)
            data = []
            
            if server_data['eintraege'].present? && server_data['eintraege'][Constants::ITEMS_KEY].present?
                for entry in server_data['eintraege'][Constants::ITEMS_KEY].compact
                    data.push({
                        :type => entry['type'],
                        Constants::DATE_KEY => entry['zeitstempel'],
                        :plan_date => entry['datum'].to_time.strftime(Constants::DATE_FORMAT),
                        :windows_user => entry['win_User'],
                        :computer_name => entry['win_Computer'],
                        :comment => entry['felder_Und_Werte'],
                        :process_name => entry['aktion']
                    })
                end
            end

            render create_json_response(:ok, THIS_SOURCE, NO_SOURCE_DETAILS, data)
        end

        # Tries to retrieve a number from a hash.
        # Defaults to given default (or 0 if no one is given), if there is no number for the given key or if the hash is nil.
        def number_from_hash_default(hash, key, default = 0)
            if hash.nil?
                return default
            end

            number = hash[key]

            if number.is_a? Numeric
                number
            else
                default
            end
        end

        # Tries to retrieve a number from a hash and put it in a hash with 'value' as key.
        def number_from_hash_default_packed(hash, key, default = 0)
            {
                Constants::VALUE_KEY => number_from_hash_default(hash, key, default)
            }
        end

        def create_fake_time_booking_data_for_date(date_time)
            first_booking_comes     = DateTime.new(date_time.year, date_time.month, date_time.day, 6, 0, 0, date_time.zone)
            first_booking_leaves    = DateTime.new(date_time.year, date_time.month, date_time.day, 9, 15, 0, date_time.zone)

            second_booking_comes   = DateTime.new(date_time.year, date_time.month, date_time.day, 9, 45, 0, date_time.zone)
            second_booking_leaves   = DateTime.new(date_time.year, date_time.month, date_time.day, 14, 0, 0, date_time.zone)

            third_booking_comes    = DateTime.new(date_time.year, date_time.month, date_time.day, 16, 0, 0, date_time.zone)
            third_booking_leaves    = DateTime.new(date_time.year, date_time.month, date_time.day, 18, 0, 0, date_time.zone)

            return [
                {
                    :time => first_booking_comes.iso8601,
                    Constants::TYPE_KEY => Constants::TIME_BOOKING_TYPE_COMES,
                    Constants::BEZEICHNUNG_KEY => "Kommt"
                },

                {
                    :time => first_booking_leaves.iso8601,
                    Constants::TYPE_KEY => Constants::TIME_BOOKING_TYPE_LEAVES,
                    Constants::BEZEICHNUNG_KEY => "Geht"
                },

                {
                    :time => second_booking_comes.iso8601,
                    Constants::TYPE_KEY => Constants::TIME_BOOKING_TYPE_COMES,
                    Constants::BEZEICHNUNG_KEY => "Kommt"
                },

                {
                    :time => second_booking_leaves.iso8601,
                    Constants::TYPE_KEY => Constants::TIME_BOOKING_TYPE_LEAVES,
                    Constants::BEZEICHNUNG_KEY => "Geht"
                },

                {
                    :time => third_booking_comes.iso8601,
                    Constants::TYPE_KEY => Constants::TIME_BOOKING_TYPE_COMES,
                    Constants::BEZEICHNUNG_KEY => "Kommt"
                },

                {
                    :time => third_booking_leaves.iso8601,
                    Constants::TYPE_KEY => Constants::TIME_BOOKING_TYPE_LEAVES,
                    Constants::BEZEICHNUNG_KEY => "Geht"
                }
            ]
        end

        def create_fake_duty_booking_data_for_date(date_time)
            time_from   = DateTime.new(date_time.year, date_time.month, date_time.day, 8, 0, 0, date_time.zone)
            time_to     = DateTime.new(date_time.year, date_time.month, date_time.day, 16, 30, 0, date_time.zone)

            token = "D"
            if date_time.day % 2 == 0
                token = "D1"
            end

            return {
                Constants::KUERZEL_KEY => token,
                Constants::DAY_FROM_KEY => time_from.iso8601,
                Constants::DAY_TO_KEY => time_to.iso8601
            }
        end

        def create_fake_breaks_booking_data_for_date(date_time)
            if date_time.day % 2 == 0
                []
            else
                break_start_1 = DateTime.new(date_time.year, date_time.month, date_time.day, 9, 0, 0, date_time.zone)
                break_end_1 = DateTime.new(date_time.year, date_time.month, date_time.day, 9, 15, 0, date_time.zone)

                break_start_2 = DateTime.new(date_time.year, date_time.month, date_time.day, 12, 0, 0, date_time.zone)
                break_end_2 = DateTime.new(date_time.year, date_time.month, date_time.day, 12, 30, 0, date_time.zone)

                break_start_3 = DateTime.new(date_time.year, date_time.month, date_time.day, 14, 0, 0, date_time.zone)
                break_end_3 = DateTime.new(date_time.year, date_time.month, date_time.day, 14, 15, 0, date_time.zone)

                break_start_4 = DateTime.new(date_time.year, date_time.month, date_time.day, 15, 0, 0, date_time.zone)
                break_end_4 = DateTime.new(date_time.year, date_time.month, date_time.day, 15, 5, 0, date_time.zone)

                [
                    {
                        Constants::TIME_BOOKING_BREAK_PLANNED_KEY => 
                        {
                            Constants::DAY_FROM_KEY => break_start_1.iso8601,
                            Constants::DAY_TO_KEY => break_end_1.iso8601
                        },

                        Constants::TIME_BOOKING_BREAK_TRACKED_KEY =>
                        {
                            Constants::DAY_FROM_KEY => break_start_1.iso8601,
                            Constants::DAY_TO_KEY => break_end_1.iso8601
                        }
                    },

                    {
                        Constants::TIME_BOOKING_BREAK_PLANNED_KEY => 
                        {
                            Constants::DAY_FROM_KEY => break_start_2.iso8601,
                            Constants::DAY_TO_KEY => break_end_2.iso8601
                        }
                    },

                    {
                        Constants::TIME_BOOKING_BREAK_PLANNED_KEY => 
                        {
                            Constants::DAY_FROM_KEY => break_start_3.iso8601,
                            Constants::DAY_TO_KEY => break_end_3.iso8601
                        }
                    },

                    {
                        Constants::TIME_BOOKING_BREAK_PLANNED_KEY => 
                        {
                            Constants::DAY_FROM_KEY => break_start_4.iso8601,
                            Constants::DAY_TO_KEY => break_end_4.iso8601
                        }
                    },
                ]
            end
        end

        def determine_header_from_employee(employee, asked_for_date)
            # Determine the current header (& month confirmation object) for the given period
            current_header = {}
            bHeaderBestaetigt = -1
            planNr = -1

            if asked_for_date.present? && employee[Constants::HEADER_LIST_KEY].present? && employee[Constants::HEADER_LIST_KEY][Constants::ITEMS_KEY].present? && employee[Constants::HEADER_LIST_KEY][Constants::ITEMS_KEY].compact.length
                for header in employee[Constants::HEADER_LIST_KEY][Constants::ITEMS_KEY].compact
                    month_from = asked_for_date.at_beginning_of_month
                    month_to = asked_for_date.end_of_month
                    if header[Constants::DT_HEADER_VON_KEY].present? && header[Constants::DT_HEADER_BIS_KEY].present? && header[Constants::DT_HEADER_VON_KEY].to_date == month_from.to_date && header[Constants::DT_HEADER_BIS_KEY].to_date == month_to.to_date
                        current_header = header
                        if header.has_key?(Constants::B_HEADER_BESTAETIGT_KEY) && header.has_key?(Constants::PLAN_NR_SERVER_KEY)
                            bHeaderBestaetigt = header[Constants::B_HEADER_BESTAETIGT_KEY] ? 1 : 0
                            planNr = header[Constants::PLAN_NR_SERVER_KEY]
                        end
                        break
                    end
                end
            end

            return current_header, bHeaderBestaetigt, planNr
        end
    end        
end
